# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pickle
from math import log

import linien_gui
import numpy as np
from linien_client.device import add_device, load_device, update_device
from linien_common.common import check_plot_data
from linien_gui.config import N_COLORS, UI_PATH, Color
from linien_gui.ui.plot_widget import INVALID_POWER
from linien_gui.ui.right_panel import RightPanel
from linien_gui.ui.spin_box import CustomDoubleSpinBox
from linien_gui.ui.sweep_control import SweepControlWidget, SweepSlider
from linien_gui.utils import color_to_hex, get_linien_app_instance, set_window_icon
from PyQt5 import QtWidgets, uic

ZOOM_STEP = 0.9
MAX_ZOOM = 50
MIN_ZOOM = 0

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def sweep_amplitude_to_zoom_step(amplitude):
    return round(log(amplitude, ZOOM_STEP))


class MainWindow(QtWidgets.QMainWindow):
    statusbar_unlocked: QtWidgets.QWidget
    signal_strenghts_unlocked: QtWidgets.QHBoxLayout
    power_channel_1: QtWidgets.QLabel
    power_channel_2: QtWidgets.QLabel
    legend_unlocked: QtWidgets.QHBoxLayout
    legend_spectrum_1: QtWidgets.QLabel
    legend_spectrum_2: QtWidgets.QLabel
    legend_spectrum_combined: QtWidgets.QLabel
    sweepControlWidget: SweepControlWidget
    sweepAmplitudeSpinBox: CustomDoubleSpinBox
    sweepCenterSpinBox: CustomDoubleSpinBox
    sweepSlider: SweepSlider
    sweepStartStopPushButton: QtWidgets.QPushButton
    top_lock_panel: QtWidgets.QWidget
    control_std: QtWidgets.QLabel
    error_std: QtWidgets.QLabel
    legend_control_signal: QtWidgets.QLabel
    legend_control_signal_history: QtWidgets.QLabel
    legend_error_signal: QtWidgets.QLabel
    legend_monitor_signal_history: QtWidgets.QLabel
    legend_slow_signal_history: QtWidgets.QLabel
    rightPanel: RightPanel
    exportParametersButton: QtWidgets.QPushButton
    importParametersButton: QtWidgets.QPushButton
    newVersionAvailableLabel: QtWidgets.QLabel
    pid_parameter_optimization_button: QtWidgets.QPushButton
    settings_toolbox: QtWidgets.QToolBox
    generalPanel: QtWidgets.QWidget
    modSpectroscopyPanel: QtWidgets.QWidget
    optimizationPanel: QtWidgets.QWidget
    loggingPanel: QtWidgets.QWidget
    viewPanel: QtWidgets.QWidget
    lockingPanel: QtWidgets.QWidget
    shutdownButton: QtWidgets.QPushButton
    closeButton: QtWidgets.QPushButton
    openDeviceManagerButton: QtWidgets.QPushButton

    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "main_window.ui", self)
        set_window_icon(self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.reset_std_history()

        # handle keyboard events
        self.setFocus()

        self.exportParametersButton.clicked.connect(self.export_parameters)
        self.importParametersButton.clicked.connect(self.import_parameters)

        def display_power(power, element):
            if power == INVALID_POWER:
                element.hide()
            else:
                element.show()
                element.setText(f"{power:.2f}")

        def display_power_channel_1(power):
            el = self.power_channel_1
            display_power(power, el)

        def display_power_channel_2(power):
            el = self.power_channel_2
            display_power(power, el)

        self.graphicsView.signal_power1.connect(display_power_channel_1)
        self.graphicsView.signal_power2.connect(display_power_channel_2)
        self.graphicsView.keyPressed.connect(self.handle_key_press)

        # by default we hide it and just show when a new version is available
        self.newVersionAvailableLabel.hide()

    def on_connection_established(self):
        self.control = self.app.control
        self.parameters = self.app.parameters

        self.parameters.lock.add_callback(self.change_sweep_control_visibility)
        self.parameters.autolock_running.add_callback(
            self.change_sweep_control_visibility
        )
        self.parameters.optimization_running.add_callback(
            self.change_sweep_control_visibility
        )

        self.parameters.to_plot.add_callback(self.update_std)

        self.parameters.pid_on_slow_enabled.add_callback(
            lambda v: self.legend_slow_signal_history.setVisible(v)
        )
        self.parameters.dual_channel.add_callback(
            lambda v: self.legend_monitor_signal_history.setVisible(not v)
        )

        self.settings_toolbox.setCurrentIndex(0)

        self.parameters.lock.add_callback(lambda *args: self.reset_std_history())

        for color_idx in range(N_COLORS):
            getattr(self.app.settings, f"plot_color_{color_idx}").add_callback(
                self.update_legend_color
            )

        self.parameters.dual_channel.add_callback(self.update_legend_text)

    def change_sweep_control_visibility(self, *args):
        al_running = self.parameters.autolock_running.value
        optimization = self.parameters.optimization_running.value
        locked = self.parameters.lock.value

        self.sweepControlWidget.setVisible(
            not al_running and not locked and not optimization
        )
        self.top_lock_panel.setVisible(locked)
        self.statusbar_unlocked.setVisible(
            not al_running and not locked and not optimization
        )

    def update_legend_color(self, *args):
        def set_color(el, color: Color):
            return el.setStyleSheet(
                "color: "
                + color_to_hex(
                    getattr(self.app.settings, f"plot_color_{color.value}").value
                )
            )

        set_color(self.legend_spectrum_1, Color.SPECTRUM1)
        set_color(self.legend_spectrum_2, Color.SPECTRUM2)
        set_color(self.legend_spectrum_combined, Color.SPECTRUM_COMBINED)
        set_color(self.legend_error_signal, Color.SPECTRUM_COMBINED)
        set_color(self.legend_control_signal, Color.CONTROL_SIGNAL)
        set_color(self.legend_control_signal_history, Color.CONTROL_SIGNAL_HISTORY)
        set_color(self.legend_slow_signal_history, Color.SLOW_HISTORY)
        set_color(self.legend_monitor_signal_history, Color.MONITOR_SIGNAL_HISTORY)

    def update_legend_text(self, dual_channel):
        self.legend_spectrum_1.setText(
            "error signal" if not dual_channel else "error signal 1"
        )
        self.legend_spectrum_2.setText(
            "monitor" if not dual_channel else "error signal 2"
        )
        self.legend_spectrum_combined.setVisible(dual_channel)

    def show(self, host: str, name: str) -> None:  # type: ignore[override]
        self.setWindowTitle(
            f"Linien spectroscopy lock {linien_gui.__version__}: {name} ({host})"
        )
        super().show()

    def closeEvent(self, *args, **kwargs):
        self.app.close_all_secondary_windows()
        super().closeEvent(*args, **kwargs)

    def show_new_version_available(self):
        self.newVersionAvailableLabel.show()

    def handle_key_press(self, key):
        logger.debug(f"key pressed {key}")

    def export_parameters(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".json"
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            f"JSON (*{default_ext})",
            options=options,
        )
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext

            try:
                add_device(self.app.client.device, path=fn)
            except KeyError:
                logger.warning(
                    f"Device with key {self.app.client.device.key} already exists in"
                    f"{fn}. Updating the device instead."
                )
                update_device(self.app.client.device, path=fn)

    def import_parameters(self):
        options = QtWidgets.QFileDialog.Options()
        fn, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            "JSON (*.json)",
            options=options,
        )
        if fn:
            try:
                self.app.client.device = load_device(
                    self.app.client.device.key, path=fn
                )
                for name, value in self.app.client.device.parameters.items():
                    param = getattr(self.app.client.parameters, name)
                    param.value = value
                self.control.exposed_write_registers()
            except KeyError:
                logger.error("Unable to load device from file. Key doesn't exist.")

    def update_std(self, to_plot, max_std_history_length=10):
        if self.parameters.lock.value and to_plot:
            to_plot = pickle.loads(to_plot)
            if to_plot and check_plot_data(True, to_plot):
                error_signal = to_plot.get("error_signal")
                control_signal = to_plot.get("control_signal")

                self.error_std_history.append(np.std(error_signal))
                self.control_std_history.append(np.std(control_signal))

                self.error_std_history = self.error_std_history[
                    -max_std_history_length:
                ]
                self.control_std_history = self.control_std_history[
                    -max_std_history_length:
                ]

                if error_signal is not None and control_signal is not None:
                    self.error_std.setText(f"{np.mean(self.error_std_history):.2f}")
                    self.control_std.setText(f"{np.mean(self.control_std_history):.2f}")

    def reset_std_history(self):
        self.error_std_history = []
        self.control_std_history = []



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import pickle
from time import time

import numpy as np
import pyqtgraph as pg
from linien_common.common import (
    DECIMATION,
    N_POINTS,
    SpectrumUncorrelatedException,
    check_plot_data,
    combine_error_signal,
    determine_shift_by_correlation,
    get_lock_point,
    get_signal_strength_from_i_q,
    update_signal_history,
)
from linien_gui.config import DEFAULT_PLOT_RATE_LIMIT, N_COLORS, Color
from linien_gui.utils import get_linien_app_instance
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal
from pyqtgraph.Qt import QtCore

# NOTE: this is required for using a pen_width > 1. There is a bug though that causes
# the plot to be way too small. Therefore, we call PlotWidget.resize() after a while
pg.setConfigOptions(
    useOpenGL=True,
    # by default, pyqtgraph tries to clean some things up using atexit. This causes
    # problems with rpyc objects as their connection is already closed. Therefore, we
    # disable this cleanup.
    exitCleanup=False,
)

# relation between counts and 1V
V = 8192

# pyqt signals enforce type, so...
INVALID_POWER = -1000


def peak_voltage_to_dBm(voltage):
    return 10 + 20 * np.log10(voltage)


class TimeXAxis(pg.AxisItem):
    """Plots x axis as time in seconds instead of point number."""

    def __init__(self, *args, parent=None, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)
        self.parent = parent
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

    def on_connection_established(self):
        # we have to wait until parameters (of parent) is available
        QtCore.QTimer.singleShot(100, self.listen_to_parameter_changes)

    def listen_to_parameter_changes(self):
        self.parent.parameters.sweep_speed.add_callback(self.force_repaint_tick_strings)
        self.parent.parameters.lock.add_callback(self.force_repaint_tick_strings)
        self.force_repaint_tick_strings()

    def force_repaint_tick_strings(self, *args):
        self.picture = None
        self.update()

    def tickStrings(self, values, scale, spacing):
        locked = self.parent.parameters.lock.value
        sweep_speed = self.parent.parameters.sweep_speed.value if not locked else 0
        time_between_points = (1 / 125e6) * 2 ** (sweep_speed) * DECIMATION
        values = [v * time_between_points for v in values]
        spacing *= time_between_points

        places = max(0, np.ceil(-np.log10(spacing * scale)))
        strings = []
        for v in values:
            vs = v * scale
            if abs(vs) < 0.001 or abs(vs) >= 10000:
                vstr = "%g" % vs
            else:
                vstr = ("%%0.%df" % places) % vs
            strings.append(vstr)
        return strings


class PlotWidget(pg.PlotWidget):
    signal_power1 = pyqtSignal(float)
    signal_power2 = pyqtSignal(float)
    keyPressed = pyqtSignal(int)

    def __init__(self, *args, **kwargs):
        super(PlotWidget, self).__init__(
            *args,
            axisItems={"bottom": TimeXAxis(parent=self, orientation="bottom")},
            **kwargs,
        )
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.getAxis("bottom").enableAutoSIPrefix(False)
        self.showGrid(x=True, y=True)
        self.setLabel("bottom", "time", units="s")

        # Causes auto-scale button (‘A’ in lower-left corner) to be hidden for this
        # PlotItem
        self.hideButtons()
        # we have our own "reset view" button instead
        self.init_reset_view_button()

        # copied from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/graphicsItems/PlotItem/PlotItem.py#L133 # noqa: E501
        # whenever something changes, we check whether to show "auto scale" button
        self.plotItem.vb.sigStateChanged.connect(
            self.check_whether_to_show_reset_view_button
        )

        # user may zoom only as far out as there is still data
        # https://stackoverflow.com/questions/18868530/pyqtgraph-limit-zoom-to-upper-lower-bound-of-axes

        self.getViewBox().setLimits(xMin=0, xMax=2048, yMin=-1, yMax=1)

        # NOTE: increasing the pen width requires OpenGL, otherwise painting gets
        # horribly slow. See: https://github.com/pyqtgraph/pyqtgraph/issues/533
        # OpenGL is enabled in the beginning of this file.
        # NOTE: OpenGL has a bug that causes the plot to be way too small. Therefore,
        # self.resize() is called below.

        self.crosshair = pg.InfiniteLine(pos=N_POINTS / 2, pen=pg.mkPen("w", width=1))
        self.addItem(self.crosshair)

        self.zero_line = pg.PlotCurveItem(pen=pg.mkPen("w", width=1))
        self.addItem(self.zero_line)
        self.signal_strength_a = pg.PlotCurveItem()
        self.addItem(self.signal_strength_a)
        self.signal_strength_a2 = pg.PlotCurveItem()
        self.addItem(self.signal_strength_a2)
        self.signal_strength_a_fill = pg.FillBetweenItem(
            self.signal_strength_a, self.signal_strength_a2
        )
        self.addItem(self.signal_strength_a_fill)
        self.signal_strength_b = pg.PlotCurveItem()
        self.addItem(self.signal_strength_b)
        self.signal_strength_b2 = pg.PlotCurveItem()
        self.addItem(self.signal_strength_b2)
        self.signal_strength_b_fill = pg.FillBetweenItem(
            self.signal_strength_b, self.signal_strength_b2
        )
        self.addItem(self.signal_strength_b_fill)
        self.signal1 = pg.PlotCurveItem()
        self.addItem(self.signal1)
        self.signal2 = pg.PlotCurveItem()
        self.addItem(self.signal2)
        self.combined_signal = pg.PlotCurveItem()
        self.addItem(self.combined_signal)

        self.control_signal = pg.PlotCurveItem()
        self.addItem(self.control_signal)
        self.control_signal_history = pg.PlotCurveItem()
        self.addItem(self.control_signal_history)
        self.slow_history = pg.PlotCurveItem()
        self.addItem(self.slow_history)
        self.monitor_signal_history = pg.PlotCurveItem()
        self.addItem(self.monitor_signal_history)

        self.zero_line.setData([0, N_POINTS - 1], [0, 0])
        self.signal1.setData([0, N_POINTS - 1], [1, 1])
        self.signal1.setData([0, N_POINTS - 1], [1, 1])
        self.combined_signal.setData([0, N_POINTS - 1], [1, 1])

        self.connection = None
        self.parameters = None
        self.last_plot_data = None
        self.plot_max = 0
        self.plot_min = np.inf
        self.touch_start = None
        self.autolock_ref_spectrum = None

        self.selection_running = False
        self.selection_boundaries = None

        self.init_overlays()
        self.init_lock_target_line()

        self._fixed_opengl_bug = False

        self.last_plot_time = 0
        self.plot_rate_limit = DEFAULT_PLOT_RATE_LIMIT

        self._plot_paused = False
        self._cached_plot_data = []
        self._should_reposition_reset_view_button = False

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        def set_pens(*args):
            pen_width = self.app.settings.plot_line_width.value

            for curve, color in {
                self.signal1: Color.SPECTRUM1,
                self.signal2: Color.SPECTRUM2,
                self.combined_signal: Color.SPECTRUM_COMBINED,
                self.control_signal: Color.CONTROL_SIGNAL,
                self.control_signal_history: Color.CONTROL_SIGNAL_HISTORY,
                self.slow_history: Color.SLOW_HISTORY,
                self.monitor_signal_history: Color.MONITOR_SIGNAL_HISTORY,
            }.items():
                r, g, b, _ = getattr(
                    self.app.settings, f"plot_color_{color.value}"
                ).value
                a = self.app.settings.plot_line_opacity.value
                curve.setPen(pg.mkPen((r, g, b, a), width=pen_width))

        for color_idx in range(N_COLORS):
            getattr(self.app.settings, f"plot_color_{color_idx}").add_callback(set_pens)
        self.app.settings.plot_line_width.add_callback(set_pens)
        self.app.settings.plot_line_opacity.add_callback(set_pens)

        self.control_signal_history_data = self.parameters.control_signal_history.value
        self.monitor_signal_history_data = self.parameters.monitor_signal_history.value

        self.parameters.to_plot.add_callback(self.replot)

        def autolock_selection_changed(value):
            if value:
                self.parameters.optimization_selection.value = False
                self.enable_area_selection(selectable_width=0.99)
                self.pause_plot_and_cache_data()
            elif not self.parameters.optimization_selection.value:
                self.disable_area_selection()
                self.resume_plot_and_clear_cache()

        self.parameters.autolock_selection.add_callback(autolock_selection_changed)

        def optimization_selection_changed(value):
            if value:
                self.parameters.autolock_selection.value = False
                self.enable_area_selection(selectable_width=0.75)
                self.pause_plot_and_cache_data()
            elif not self.parameters.autolock_selection.value:
                self.disable_area_selection()
                self.resume_plot_and_clear_cache()

        self.parameters.optimization_selection.add_callback(
            optimization_selection_changed
        )

        def show_or_hide_crosshair(automatic_mode):
            self.crosshair.setVisible(not automatic_mode)

        self.parameters.automatic_mode.add_callback(show_or_hide_crosshair)

    def _to_data_coords(self, event):
        pos = self.plotItem.vb.mapSceneToView(event.pos())
        x, y = pos.x(), pos.y()
        return x, y

    def mouseMoveEvent(self, event):
        if not self.selection_running:
            super().mouseMoveEvent(event)
        else:
            if self.touch_start is None:
                return

            x0, y0 = self.touch_start

            x, y = self._to_data_coords(event)
            x = self._within_boundaries(x)
            self.set_selection_overlay(x0, x - x0)

    def init_overlays(self):
        self.overlay = pg.LinearRegionItem(values=(0, 0), movable=False)
        self.overlay.setVisible(False)
        self.addItem(self.overlay)

        self.boundary_overlays = [
            pg.LinearRegionItem(
                values=(0, 0),
                movable=False,
                brush=(0, 0, 0, 200),
            )
            for i in range(2)
        ]
        for i, overlay in enumerate(self.boundary_overlays):
            overlay.setVisible(False)
            # make outer borders invisible, see
            # https://github.com/pyqtgraph/pyqtgraph/issues/462
            overlay.lines[i].setPen((0, 0, 0, 0))
            self.addItem(overlay)

    def init_lock_target_line(self):
        self.lock_target_line = pg.InfiniteLine(movable=False)
        self.lock_target_line.setValue(1000)
        self.addItem(self.lock_target_line)

    def set_selection_overlay(self, x_start, width):
        self.overlay.setRegion((x_start, x_start + width))

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        if self.selection_running:
            if event.button() == QtCore.Qt.RightButton:
                return

            x, y = self._to_data_coords(event)

            if self.selection_running:
                if x < self.selection_boundaries[0] or x > self.selection_boundaries[1]:
                    return

            self.touch_start = x, y
            self.set_selection_overlay(x, 0)
            self.overlay.setVisible(True)

    def _within_boundaries(self, x):
        boundaries = (
            self.selection_boundaries if self.selection_running else [0, N_POINTS]
        )

        if x < boundaries[0]:
            return boundaries[0]
        if x > boundaries[1]:
            return boundaries[1]
        return x

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)

        if self.selection_running:
            if self.touch_start is None:
                return

            x, y = self._to_data_coords(event)
            x = self._within_boundaries(x)
            x0, y0 = self.touch_start
            xdiff = np.abs(x0 - x)
            xmax = len(self.last_plot_data[0]) - 1
            if xdiff / xmax < 0.01:
                # it was a click
                pass
            else:
                # it was a selection
                if self.selection_running:
                    if self.parameters.autolock_selection.value:
                        last_combined_error_signal = self.last_plot_data[2]
                        self.parameters.autolock_selection.value = False

                        self.control.start_autolock(
                            # we pickle it here because otherwise a netref is
                            # transmitted which blocks the autolock
                            *sorted([x0, x]),
                            pickle.dumps(last_combined_error_signal),
                            additional_spectra=pickle.dumps(self._cached_plot_data),
                        )

                        (
                            mean_signal,
                            target_slope_rising,
                            target_zoom,
                            rolled_error_signal,
                            line_width,
                            peak_idxs,
                        ) = get_lock_point(
                            last_combined_error_signal, *sorted((int(x0), int(x)))
                        )
                        self.autolock_ref_spectrum = rolled_error_signal
                    elif self.parameters.optimization_selection.value:
                        dual_channel = self.parameters.dual_channel.value
                        channel = self.parameters.optimization_channel.value
                        spectrum = self.last_plot_data[
                            0 if not dual_channel else (0, 1)[channel]
                        ]
                        self.parameters.optimization_selection.value = False
                        points = sorted([int(x0), int(x)])
                        self.control.start_optimization(*points, pickle.dumps(spectrum))

            self.overlay.setVisible(False)
            self.touch_start = None

    def replot(self, to_plot):
        time_beginning = time()

        if self._should_reposition_reset_view_button:
            self._should_reposition_reset_view_button = False
            self.position_reset_view_button()

        if (
            time_beginning - self.last_plot_time <= self.plot_rate_limit
            and not self._plot_paused
        ):
            # don't plot too often as it only causes unnecessary load this does not
            # apply if plot is paused, because in this case we want to collect all the
            # data that we can get in order to pass it to the autolock
            return

        self.last_plot_time = time_beginning

        # NOTE: this is necessary if OpenGL is activated. Otherwise, the plot is way too
        # small. This command apparently causes a repaint and works fine even though the
        # values are nonsense.
        if not self._fixed_opengl_bug:
            self._fixed_opengl_bug = True
            self.resize(
                self.parent().frameGeometry().width(),
                self.parent().frameGeometry().height(),
            )

        if self.parameters.pause_acquisition.value:
            return

        if to_plot is not None:
            to_plot = pickle.loads(to_plot)

            if to_plot is None:
                return

            if not check_plot_data(self.parameters.lock.value, to_plot):
                return

            # we also call this if the laser is not locked because it resets the history
            # in this case
            history, slow_history = self.update_signal_history(to_plot)

            if self.parameters.lock.value:
                self.signal1.setVisible(False)
                self.signal2.setVisible(False)
                self.control_signal.setVisible(True)
                self.control_signal_history.setVisible(True)
                self.slow_history.setVisible(self.parameters.pid_on_slow_enabled.value)
                self.monitor_signal_history.setVisible(
                    not self.parameters.dual_channel.value
                )
                self.combined_signal.setVisible(True)
                self.signal_strength_a.setVisible(False)
                self.signal_strength_b.setVisible(False)
                self.signal_strength_a2.setVisible(False)
                self.signal_strength_b2.setVisible(False)
                self.signal_strength_a_fill.setVisible(False)
                self.signal_strength_b_fill.setVisible(False)

                error_signal, control_signal = (
                    to_plot["error_signal"],
                    to_plot["control_signal"],
                )
                all_signals = (error_signal, control_signal, history, slow_history)

                self.plot_data_locked(to_plot)
                self.plot_autolock_target_line(None)
            else:
                dual_channel = self.parameters.dual_channel.value
                self.signal1.setVisible(True)
                monitor_signal = to_plot.get("monitor_signal")
                error_signal_2 = to_plot.get("error_signal_2")
                self.signal2.setVisible(
                    error_signal_2 is not None or monitor_signal is not None
                )
                self.combined_signal.setVisible(dual_channel)
                self.control_signal.setVisible(False)
                self.control_signal_history.setVisible(False)
                self.slow_history.setVisible(False)
                self.monitor_signal_history.setVisible(False)

                s1 = to_plot["error_signal_1"]
                s2 = error_signal_2 if error_signal_2 is not None else monitor_signal

                combined_error_signal = combine_error_signal(
                    (s1, s2),
                    dual_channel,
                    self.parameters.channel_mixing.value,
                    self.parameters.combined_offset.value,
                )

                if self._plot_paused:
                    self._cached_plot_data.append(combined_error_signal)
                    # don't save too much
                    self._cached_plot_data = self._cached_plot_data[-20:]
                    return

                all_signals = [s1, s2] + [combined_error_signal]
                self.last_plot_data = all_signals

                self.plot_data_unlocked((s1, s2), combined_error_signal)
                self.plot_autolock_target_line(combined_error_signal)

                if (self.parameters.modulation_frequency.value != 0) and (
                    not self.parameters.pid_only_mode.value
                ):
                    # check whether to plot signal strengths using quadratures
                    s1q = to_plot.get("error_signal_1_quadrature")
                    s2q = to_plot.get("error_signal_2_quadrature")

                    self.signal_strength_a.setVisible(s1q is not None)
                    self.signal_strength_a2.setVisible(s1q is not None)
                    self.signal_strength_a_fill.setVisible(s1q is not None)

                    self.signal_strength_b.setVisible(s2q is not None)
                    self.signal_strength_b2.setVisible(s2q is not None)
                    self.signal_strength_b_fill.setVisible(s2q is not None)

                    if s1q is not None:
                        max_signal_strength_V = (
                            self.plot_signal_strength(
                                s1,
                                s1q,
                                self.signal_strength_a,
                                self.signal_strength_a2,
                                self.signal_strength_a_fill,
                                self.parameters.offset_a.value,
                                self.app.settings.plot_color_0.value,
                            )
                            / V
                        )
                        all_signals.append(
                            [
                                max_signal_strength_V * V,
                                -1 * max_signal_strength_V * V,
                            ]
                        )

                        self.signal_power1.emit(
                            peak_voltage_to_dBm(max_signal_strength_V)
                        )
                    else:
                        self.signal_power1.emit(INVALID_POWER)

                    if s2q is not None:
                        max_signal_strength2_V = (
                            self.plot_signal_strength(
                                s2,
                                s2q,
                                self.signal_strength_b,
                                self.signal_strength_b2,
                                self.signal_strength_b_fill,
                                self.parameters.offset_b.value,
                                self.app.settings.plot_color_1.value,
                            )
                            / V
                        )

                        all_signals.append(
                            [
                                max_signal_strength2_V * V,
                                -1 * max_signal_strength2_V * V,
                            ]
                        )

                        self.signal_power2.emit(
                            peak_voltage_to_dBm(max_signal_strength2_V)
                        )
                    else:
                        self.signal_power2.emit(INVALID_POWER)
                else:
                    self.signal_strength_a.setVisible(False)
                    self.signal_strength_b.setVisible(False)
                    self.signal_strength_a2.setVisible(False)
                    self.signal_strength_b2.setVisible(False)
                    self.signal_strength_a_fill.setVisible(False)
                    self.signal_strength_b_fill.setVisible(False)

                    self.signal_power1.emit(INVALID_POWER)
                    self.signal_power2.emit(INVALID_POWER)

        time_end = time()
        time_diff = time_end - time_beginning
        new_rate_limit = 2 * time_diff

        if new_rate_limit < DEFAULT_PLOT_RATE_LIMIT:
            new_rate_limit = DEFAULT_PLOT_RATE_LIMIT

        self.plot_rate_limit = new_rate_limit

    def plot_signal_strength(
        self, i, q, signal, neg_signal, fill, channel_offset, color
    ):
        # we have to subtract channel offset here and will add it back in the end
        i -= int(round(channel_offset))
        q -= int(round(channel_offset))
        signal_strength = get_signal_strength_from_i_q(i, q)

        r, g, b, *stuff = color

        x = list(range(len(signal_strength)))
        signal_strength_scaled = signal_strength / V
        upper = (channel_offset / V) + signal_strength_scaled
        lower = (channel_offset / V) - 1 * signal_strength_scaled

        brush = pg.mkBrush(r, g, b, self.app.settings.plot_fill_opacity.value)
        fill.setBrush(brush)

        invisible_pen = pg.mkPen("k", width=0.00001)
        signal.setData(x, upper, pen=invisible_pen)
        neg_signal.setData(x, lower, pen=invisible_pen)
        return np.max([np.max(upper), -1 * np.min(lower)]) * V

    def plot_data_unlocked(self, error_signals, combined_signal):
        error_signal1, error_signal2 = error_signals
        self.signal1.setData(list(range(len(error_signal1))), error_signal1 / V)
        self.signal2.setData(list(range(len(error_signal2))), error_signal2 / V)
        self.combined_signal.setData(
            list(range(len(combined_signal))), combined_signal / V
        )

    def plot_data_locked(self, signals):
        error_signal = signals["error_signal"]
        control_signal = signals["control_signal"]
        self.combined_signal.setData(list(range(len(error_signal))), error_signal / V)
        self.control_signal.setData(list(range(len(error_signal))), control_signal / V)

    def plot_autolock_target_line(self, combined_error_signal):
        if (
            self.autolock_ref_spectrum is not None
            and self.parameters.autolock_preparing.value
        ):
            sweep_amplitude = self.parameters.sweep_amplitude.value
            zoom_factor = 1 / sweep_amplitude
            initial_zoom_factor = (
                1 / self.parameters.autolock_initial_sweep_amplitude.value
            )

            try:
                shift, _1, _2 = determine_shift_by_correlation(
                    zoom_factor / initial_zoom_factor,
                    self.autolock_ref_spectrum,
                    combined_error_signal,
                )
                shift *= zoom_factor / initial_zoom_factor
                length = len(combined_error_signal)
                shift = (length / 2) - (shift / 2 * length)

                self.lock_target_line.setVisible(True)
                self.lock_target_line.setValue(shift)

            except SpectrumUncorrelatedException:
                self.lock_target_line.setVisible(False)
        else:
            self.lock_target_line.setVisible(False)

    def keyPressEvent(self, event):
        # we listen here in addition to the main window because some events are only
        # caught here
        self.keyPressed.emit(event.key())

    def update_signal_history(self, to_plot):
        update_signal_history(
            self.control_signal_history_data,
            self.monitor_signal_history_data,
            to_plot,
            self.parameters.lock.value,
            self.parameters.control_signal_history_length.value,
        )

        if self.parameters.lock.value:
            x_axis_length = N_POINTS

            def scale(arr):
                timescale = self.parameters.control_signal_history_length.value
                if arr:
                    arr = np.array(arr)
                    arr -= arr[0]
                    arr *= 1 / timescale * x_axis_length
                return arr

            history = self.control_signal_history_data["values"]
            self.control_signal_history.setData(
                scale(self.control_signal_history_data["times"]), np.array(history) / V
            )

            slow_values = self.control_signal_history_data["slow_values"]
            self.slow_history.setData(
                scale(self.control_signal_history_data["slow_times"]),
                np.array(slow_values) / V,
            )

            if not self.parameters.dual_channel.value:
                self.monitor_signal_history.setData(
                    scale(self.monitor_signal_history_data["times"]),
                    np.array(self.monitor_signal_history_data["values"]) / V,
                )

            return history, slow_values
        return [], []

    def enable_area_selection(self, selectable_width=0.5):
        self.selection_running = True

        # there are some glitches if the width of the overlay is exactly right.
        # Therefore we make it a little wider.
        extra_width = N_POINTS / 100
        x_axis_length = N_POINTS
        boundary_width = (x_axis_length * (1 - selectable_width)) / 2.0

        self.selection_boundaries = (boundary_width, x_axis_length - boundary_width)

        self.boundary_overlays[0].setRegion((-extra_width, boundary_width))
        self.boundary_overlays[1].setRegion(
            (x_axis_length - boundary_width, x_axis_length + extra_width)
        )

        for overlay in self.boundary_overlays:
            overlay.setVisible(True)

    def disable_area_selection(self):
        self.selection_running = False

        for overlay in self.boundary_overlays:
            overlay.setVisible(False)

    def pause_plot_and_cache_data(self):
        """This function pauses plot updates. All incoming data is cached though.
        This is useful for letting the user select a line that is then used in
        the autolock."""
        self._plot_paused = True

    def resume_plot_and_clear_cache(self):
        """Resumes plotting again."""
        self._plot_paused = False
        self._cached_plot_data = []

    def init_reset_view_button(self):
        self.reset_view_button = QtWidgets.QPushButton(self)
        self.reset_view_button.setText("Reset view")
        self.reset_view_button.setStyleSheet("padding: 10px; font-weight: bold")
        icon = QtGui.QIcon.fromTheme("view-restore")
        self.reset_view_button.setIcon(icon)
        self.reset_view_button.clicked.connect(self.reset_view)
        self.position_reset_view_button()

    # called when widget is resized
    def position_reset_view_button(self):
        pos = QtCore.QPoint(
            self.geometry().width() - self.reset_view_button.geometry().width() - 25,
            25,
        )
        self.reset_view_button.move(pos)

    def check_whether_to_show_reset_view_button(self):
        # copied from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/graphicsItems/PlotItem/PlotItem.py#L1195 # noqa: E501
        auto_scale_disabled = not all(self.plotItem.vb.autoRangeEnabled())
        if auto_scale_disabled:
            self.reset_view_button.show()
        else:
            self.reset_view_button.hide()

    def reset_view(self):
        self.enableAutoRange()

    def resizeEvent(self, event, *args, **kwargs):
        super().resizeEvent(event, *args, **kwargs)

        # we don't do it directly here because this causes problems for some reason
        self._should_reposition_reset_view_button = True



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_gui.utils import get_linien_app_instance, param2ui
from PyQt5 import QtCore, QtWidgets


class LockStatusPanel(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(LockStatusPanel, self).__init__(*args, **kwargs)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)
        QtCore.QTimer.singleShot(100, self.ready)

    def ready(self):
        self.parent = self.parent()
        self.parent.stopLockPushButton.clicked.connect(self.on_stop_lock)
        self.parent.controlSignalHistoryLengthSpinBox.setKeyboardTracking(False)
        self.parent.controlSignalHistoryLengthSpinBox.valueChanged.connect(
            self.on_control_signal_history_length_changed
        )

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        def update_status(_):
            locked = self.parameters.lock.value
            task = self.parameters.task.value
            al_failed = self.parameters.autolock_failed.value
            running = self.parameters.autolock_running.value
            retrying = self.parameters.autolock_retrying.value
            percentage = self.parameters.autolock_percentage.value
            preparing = self.parameters.autolock_preparing.value

            if locked or (task is not None and not al_failed):
                self.show()
            else:
                self.hide()

            if task:
                watching = self.parameters.autolock_watching.value
            else:
                running = False
                watching = False

            def set_text(text):
                self.parent.lock_status.setText(text)

            if not running and locked:
                set_text("Locked!")
            if running and watching:
                set_text("Locked! Watching continuously...")
            if running and not watching and not locked and preparing:
                if not retrying:
                    set_text(f"Autolock is running... Analyzing data ({percentage} %%)")
                else:
                    set_text("Trying again to lock...")

        for param in (
            self.parameters.lock,
            self.parameters.task,
            self.parameters.autolock_running,
            self.parameters.autolock_preparing,
            self.parameters.autolock_watching,
            self.parameters.autolock_failed,
            self.parameters.autolock_locked,
            self.parameters.autolock_retrying,
            self.parameters.autolock_percentage,
        ):
            param.add_callback(update_status)

        param2ui(
            self.parameters.control_signal_history_length,
            self.parent.controlSignalHistoryLengthSpinBox,
        )

    def on_stop_lock(self):
        self.parameters.fetch_additional_signals.value = True

        if self.parameters.task.value is not None:
            # this may be autolock or psd acquisition
            self.parameters.task.value.stop()
            self.parameters.task.value = None

        self.control.exposed_start_sweep()

    def on_control_signal_history_length_changed(self):
        self.parameters.control_signal_history_length.value = (
            self.parent.controlSignalHistoryLengthSpinBox.value()
        )



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import TYPE_CHECKING, Any, Callable, Tuple

from linien_client.remote_parameters import RemoteParameter
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QRadioButton,
    QSlider,
    QSpinBox,
    QTabWidget,
)
from pyqtgraph.Qt import QtGui

if TYPE_CHECKING:
    from linien_gui.app import LinienApp


def get_linien_app_instance() -> "LinienApp":
    return QtWidgets.QApplication.instance()  # type: ignore[return-value]


def param2ui(
    parameter: RemoteParameter,
    element: QtWidgets.QWidget,
    process_value: Callable[[Any], Any] = lambda x: x,
):
    """
    Updates ui elements according to parameter values.

    Listens to parameter changes and sets the value of `element` automatically.
    Optionally, the value can be processed using `process_value`. This function should
    be used because it automatically blocks signal emission from the target element;
    otherwise this can cause nasty endless loops when quickly changing a parameter
    multiple times.
    """

    def on_change(value: Any, element=element) -> None:
        element.blockSignals(True)

        value = process_value(value)

        if isinstance(element, (QSlider, QSpinBox, QDoubleSpinBox)):
            element.setValue(value)
        elif isinstance(element, (QCheckBox, QRadioButton)):
            element.setChecked(value)
        elif isinstance(element, (QTabWidget, QComboBox)):
            element.setCurrentIndex(int(value))
        else:
            raise TypeError(f"Unsupported element type {type(element)}")

        element.blockSignals(False)

    parameter.add_callback(on_change)


def set_window_icon(window: QtWidgets.QMainWindow) -> None:
    icon_name = os.path.join(*os.path.split(__file__)[:-1], "icon.ico")
    window.setWindowIcon(QtGui.QIcon(icon_name))


def color_to_hex(color: Tuple[int, int, int]) -> str:
    result = ""
    for part_idx in range(3):
        result += ("00" + hex(color[part_idx]).lstrip("0x"))[-2:]

    return "#" + result


def hex_to_color(hex_: str) -> Tuple[int, ...]:
    hex_ = hex_.lstrip("#")
    return tuple(int(hex_[i : i + 2], 16) for i in (0, 2, 4))


class RandomColorChoser:
    def __init__(self):
        self.index = 0
        # pick one to turn into an actual colormap generated using
        # ttps://mokole.com/palette.html and shuffled using random.shuffle
        self.lut = [
            "#2e8b57",
            "#0000ff",
            "#87cefa",
            "#ff1493",
            "#adff2f",
            "#b03060",
            "#6495ed",
            "#90ee90",
            "#dc143c",
            "#ffff00",
            "#483d8b",
            "#f08080",
            "#8b4513",
            "#00ff00",
            "#da70d6",
            "#f4a460",
            "#008000",
            "#00ff7f",
            "#808000",
            "#7b68ee",
            "#a9a9a9",
            "#ff8c00",
            "#00008b",
            "#f0e68c",
            "#ff0000",
            "#800080",
            "#2f4f4f",
            "#ff00ff",
            "#00ffff",
            "#8a2be2",
        ]

    def get(self):
        color = self.lut[self.index]
        if self.index < len(self.lut) - 1:
            self.index += 1
        else:
            self.index = 0

        return hex_to_color(color)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Tuple

import linien_gui
from linien_client.device import Device, delete_device, load_device_list, move_device
from linien_common.communication import RestorableParameterValues
from linien_gui.config import UI_PATH
from linien_gui.dialogs import (
    LoadingDialog,
    ask_for_parameter_restore_dialog,
    error_dialog,
    question_dialog,
    show_installation_progress_widget,
)
from linien_gui.threads import ConnectionThread
from linien_gui.ui.new_device_dialog import NewDeviceDialog
from linien_gui.utils import get_linien_app_instance, set_window_icon
from PyQt5 import QtCore, QtWidgets, uic
from PyQt5.QtWidgets import QListWidget, QPushButton


class DeviceManager(QtWidgets.QMainWindow):
    addButton: QPushButton
    connectButton: QPushButton
    deviceList: QListWidget
    editButton: QPushButton
    moveDownButton: QPushButton
    moveUpButton: QPushButton
    removeButton: QPushButton

    def __init__(self, *args, **kwargs):
        super(DeviceManager, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "device_manager.ui", self)
        self.setWindowTitle(f"Linien spectroscopy lock v{linien_gui.__version__}")
        set_window_icon(self)
        self.app = get_linien_app_instance()
        QtCore.QTimer.singleShot(100, lambda: self.populate_device_list())

    def keyPressEvent(self, event):
        key = event.key()

        if key in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            self.connect()

    def populate_device_list(self):
        self.devices = load_device_list()

        self.deviceList.clear()
        for device in self.devices:
            self.deviceList.addItem(f"{device.name} ({device.host})")

    def connect(self) -> None:
        if not self.devices:
            return
        else:
            self.connect_to_device(self.devices[self.get_list_index()])

    def connect_to_device(self, device: Device):
        loading_dialog = LoadingDialog(self, device.host)
        loading_dialog.show()

        aborted = {}

        self.connection_thread = ConnectionThread(device)

        def was_aborted(*args):
            aborted["aborted"] = True

        loading_dialog.aborted.connect(was_aborted)

        # Define slot functions to be connected ----------------------------------------
        def on_client_connected(client):
            loading_dialog.hide()
            if not aborted:
                self.app.client_connected(client)

        def handle_server_not_installed():
            loading_dialog.hide()
            if not aborted:
                display_question = (
                    "The server is not yet installed on the device. Should it be "
                    "installed? (Requires internet connection on RedPitaya)"
                )
                if question_dialog(self, display_question, "Install server?"):
                    show_installation_progress_widget(
                        parent=self,
                        device=device,
                        callback=lambda: self.connect_to_device(device),
                    )

        def handle_invalid_server_version(
            remote_version: str, client_version: str
        ) -> None:
            loading_dialog.hide()
            if not aborted:
                display_question = (
                    f"Server version ({remote_version}) does not match the client "
                    f"({client_version}) version. Should the corresponding server "
                    f"version be installed?"
                )
                if question_dialog(
                    self, display_question, "Install corresponding version?"
                ):
                    show_installation_progress_widget(
                        parent=self,
                        device=device,
                        callback=lambda: self.connect_to_device(device),
                    )

        def handle_authentication_exception():
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    "Error at authentication.\n"
                    "Check username and password (by default both are 'root')  and "
                    "verify that you don't have any offending SSH keys in your known "
                    "hosts file."
                )
                error_dialog(self, display_error)

        def handle_general_connection_error():
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    "Unable to connect to device. If you are connecting by hostname "
                    "(i.e. `rp-xxxxxx.local`), try using IP address instead."
                )
                error_dialog(self, display_error)

        def handle_other_exception(exception):
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    f"Exception occurred when connecting to the device:\n\n {exception}"
                )
                error_dialog(self, display_error)

        def ask_for_parameter_restore(
            parameter_difference: Dict[
                str, Tuple[RestorableParameterValues, RestorableParameterValues]
            ]
        ) -> None:
            question = (
                "Linien on RedPitaya is running with different parameters than the "
                "ones saved locally on this machine. Do you want to upload the local "
                "parameters or keep the remote ones? Note that remote parameters are "
                "only saved if Linien server was shut down properly, not when "
                "unplugging the power plug. In this case, you should update your local "
                "parameters."
            )
            should_restore = ask_for_parameter_restore_dialog(
                self, question, "Restore parameters?"
            )
            if should_restore:
                self.connection_thread.restore_parameters(parameter_difference)
            self.connection_thread.add_callbacks_to_write_parameters_to_disk_on_change()

        def handle_connection_lost():
            error_dialog(self, "Lost connection to the server!")
            self.app.quit()

        # Connect slots to signals -----------------------------------------------------
        self.connection_thread.client_connected.connect(on_client_connected)

        self.connection_thread.server_not_installed_exception_raised.connect(
            handle_server_not_installed
        )
        self.connection_thread.invalid_server_version_exception_raised.connect(
            handle_invalid_server_version
        )
        self.connection_thread.authentication_exception_raised.connect(
            handle_authentication_exception
        )
        self.connection_thread.general_connection_exception_raised.connect(
            handle_general_connection_error
        )
        self.connection_thread.other_exception_raised.connect(handle_other_exception)
        self.connection_thread.parameter_difference.connect(ask_for_parameter_restore)
        self.connection_thread.connection_lost.connect(handle_connection_lost)

        # Start the worker -------------------------------------------------------------
        self.connection_thread.start()

    def get_list_index(self):
        """Get the currently selected device index from the device list."""
        return self.deviceList.currentIndex().row()

    def reload_device_data(self) -> None:
        # not very elegant...
        QtCore.QTimer.singleShot(100, self.populate_device_list)

    def new_device(self) -> None:
        """Open the dialog to create a new device."""
        self.dialog = NewDeviceDialog()
        self.dialog.setModal(True)
        self.dialog.show()
        self.dialog.accepted.connect(self.reload_device_data)

    def edit_device(self) -> None:
        """Open the dialog to edit the currently selected device."""
        device = self.devices[self.get_list_index()]

        self.dialog = NewDeviceDialog(device)
        self.dialog.setModal(True)
        self.dialog.show()
        self.dialog.accepted.connect(self.reload_device_data)

    def move_device_up(self) -> None:
        """Move the currently selected device up in the list."""
        self.move_device_in_list(-1)

    def move_device_down(self) -> None:
        """Move the currently selected device down in the list."""
        self.move_device_in_list(1)

    def move_device_in_list(self, direction: int) -> None:
        """Move the currently selected device in the list by the given direction."""
        selected_index = self.get_list_index()
        selected_device = self.devices[selected_index]
        move_device(selected_device, direction)
        self.populate_device_list()
        self.deviceList.setCurrentRow(selected_index + direction)

    def remove_device(self) -> None:
        """
        Remove the currently selected device from the list and save new list to disk.
        """
        selected_device = self.devices[self.get_list_index()]
        delete_device(selected_device)
        self.populate_device_list()

    def selected_device_changed(self) -> None:
        disable_buttons = True

        if self.get_list_index() >= 0:
            if self.devices:
                disable_buttons = False

        for button in [
            self.connectButton,
            self.removeButton,
            self.editButton,
            self.moveUpButton,
            self.moveDownButton,
        ]:
            button.setEnabled(not disable_buttons)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

from linien_client.remote_parameters import RemoteParameter
from linien_common.common import FilterType
from linien_gui.config import UI_PATH
from linien_gui.ui.spin_box import CustomDoubleSpinBox, CustomDoubleSpinBoxNoSign
from linien_gui.utils import get_linien_app_instance, param2ui
from PyQt5 import QtWidgets, uic


class SpectroscopyPanel(QtWidgets.QWidget):
    CHANNEL: Union[str, None] = None

    automaticFilterCheckBox: QtWidgets.QCheckBox
    demodulationFrequencyComboBox: CustomDoubleSpinBoxNoSign
    demodulationPhaseSpinBox: CustomDoubleSpinBoxNoSign
    invertCheckBox: QtWidgets.QCheckBox
    manualFilterWidget: QtWidgets.QWidget
    signalOffsetSpinBox: CustomDoubleSpinBox

    def __init__(self, *args):
        super(SpectroscopyPanel, self).__init__(*args)
        uic.loadUi(UI_PATH / "spectroscopy_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        def change_filter_frequency(filter_i):
            self.get_param(f"filter_{filter_i}_frequency").value = getattr(
                self, f"filter_{filter_i}_frequency"
            ).value()
            self.control.exposed_write_registers()

        def change_filter_enabled(filter_i):
            filter_enabled = int(
                getattr(self, f"filter_{filter_i}_enabled").checkState() > 0
            )
            self.get_param(f"filter_{filter_i}_enabled").value = filter_enabled
            self.control.exposed_write_registers()

        def change_filter_type(filter_i):
            param = self.get_param(f"filter_{filter_i}_type")
            current_idx = getattr(self, f"filter_{filter_i}_type").currentIndex()
            param.value = (FilterType.LOW_PASS, FilterType.HIGH_PASS)[current_idx]
            self.control.exposed_write_registers()

        for filter_i in [1, 2]:
            for key, fct in {
                f"change_filter_{filter_i}_frequency": change_filter_frequency,
                f"change_filter_{filter_i}_enabled": change_filter_enabled,
                f"change_filter_{filter_i}_type": change_filter_type,
            }.items():
                setattr(
                    self,
                    key,
                    lambda *args, fct=fct, filter_i=filter_i: fct(filter_i),
                )

        self.signalOffsetSpinBox.setKeyboardTracking(False)
        self.signalOffsetSpinBox.valueChanged.connect(self.change_signal_offset)
        self.demodulationPhaseSpinBox.setKeyboardTracking(False)
        self.demodulationPhaseSpinBox.valueChanged.connect(self.change_demod_phase)
        self.demodulationFrequencyComboBox.currentIndexChanged.connect(
            self.change_demod_multiplier
        )

        for filter_i in [1, 2]:

            def get_(parent, attr, filter_i=filter_i):
                return getattr(parent, attr.format(filter_i))

            get_(self, "filter_{}_enabled").stateChanged.connect(
                get_(self, "change_filter_{}_enabled")
            )
            freq_input = get_(self, "filter_{}_frequency")
            freq_input.setKeyboardTracking(False)
            freq_input.valueChanged.connect(get_(self, "change_filter_{}_frequency"))
            get_(self, "filter_{}_type").currentIndexChanged.connect(
                get_(self, "change_filter_{}_type")
            )

        def automatic_changed(value):
            self.get_param("filter_automatic").value = value
            self.control.exposed_write_registers()

        self.automaticFilterCheckBox.stateChanged.connect(automatic_changed)

        def invert_changed(value):
            self.get_param("invert").value = bool(value)
            self.control.exposed_write_registers()

        self.invertCheckBox.stateChanged.connect(invert_changed)

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.get_param("demodulation_phase"), self.demodulationPhaseSpinBox)
        param2ui(
            self.get_param("demodulation_multiplier"),
            self.demodulationFrequencyComboBox,
            lambda value: value - 1,
        )
        param2ui(
            self.get_param("offset"), self.signalOffsetSpinBox, lambda v: v / 8191.0
        )
        param2ui(self.get_param("invert"), self.invertCheckBox)
        param2ui(self.get_param("filter_automatic"), self.automaticFilterCheckBox)

        def filter_automatic_changed(value):
            self.manualFilterWidget.setVisible(value)
            self.manualFilterWidget.setVisible(not value)

        self.get_param("filter_automatic").add_callback(filter_automatic_changed)

        for filter_i in [1, 2]:
            param2ui(
                self.get_param(f"filter_{filter_i}_enabled"),
                getattr(self, f"filter_{filter_i}_enabled"),
            )
            param2ui(
                self.get_param(f"filter_{filter_i}_frequency"),
                getattr(self, f"filter_{filter_i}_frequency"),
            )
            param2ui(
                self.get_param(f"filter_{filter_i}_type"),
                getattr(self, f"filter_{filter_i}_type"),
                lambda type_: {FilterType.LOW_PASS: 0, FilterType.HIGH_PASS: 1}[type_],
            )

    def get_param(self, name: str) -> RemoteParameter:
        return getattr(self.parameters, f"{name}_{self.CHANNEL}")

    def change_signal_offset(self):
        self.get_param("offset").value = self.signalOffsetSpinBox.value() * 8191
        self.control.exposed_write_registers()

    def change_demod_phase(self):
        self.get_param("demodulation_phase").value = (
            self.demodulationPhaseSpinBox.value()
        )
        self.control.exposed_write_registers()

    def change_demod_multiplier(self, idx):
        self.get_param("demodulation_multiplier").value = idx + 1
        self.control.exposed_write_registers()


class SpectroscopyChannelAPanel(SpectroscopyPanel):
    CHANNEL = "a"


class SpectroscopyChannelBPanel(SpectroscopyPanel):
    CHANNEL = "b"



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>

#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from enum import Enum
from pathlib import Path
from typing import Callable, Iterator, Tuple

from linien_common.config import USER_DATA_PATH, create_backup_file

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

UI_PATH = Path(__file__).parents[0].resolve() / "ui"
SETTINGS_STORE_FILENAME = "settings.json"
# don't plot more often than once per `DEFAULT_PLOT_RATE_LIMIT` seconds
DEFAULT_PLOT_RATE_LIMIT = 0.1

DEFAULT_COLORS = [
    (200, 0, 0, 200),
    (0, 200, 0, 200),
    (0, 0, 200, 200),
    (200, 200, 0, 200),
    (200, 0, 200, 200),
]
N_COLORS = len(DEFAULT_COLORS)


class Color(Enum):
    SPECTRUM1 = 0
    SPECTRUM2 = 1
    SPECTRUM_COMBINED = 2
    CONTROL_SIGNAL = 0
    CONTROL_SIGNAL_HISTORY = 1
    SLOW_HISTORY = 3
    MONITOR_SIGNAL_HISTORY = 4


class Setting:
    def __init__(
        self,
        min_=None,
        max_=None,
        start=None,
    ):
        self.min = min_
        self.max = max_
        self._value = start
        self.start = start
        self._callbacks = set()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.min is not None and value < self.min:
            value = self.min
        if self.max is not None and value > self.max:
            value = self.max
        self._value = value

        # We copy it because a listener could remove a listener --> this would cause an
        # error in this loop.
        for callback in self._callbacks.copy():
            callback(value)

    def add_callback(self, function: Callable, call_immediatly: bool = True):
        self._callbacks.add(function)

        if call_immediatly:
            if self._value is not None:
                function(self._value)

    def remove_callback(self, function):
        if function in self._callbacks:
            self._callbacks.remove(function)


class Settings:
    def __init__(self):
        self.plot_line_width = Setting(start=2, min_=0.1, max_=100)
        self.plot_line_opacity = Setting(start=230, min_=0, max_=255)
        self.plot_fill_opacity = Setting(start=70, min_=0, max_=255)
        self.plot_color_0 = Setting(start=DEFAULT_COLORS[0])
        self.plot_color_1 = Setting(start=DEFAULT_COLORS[1])
        self.plot_color_2 = Setting(start=DEFAULT_COLORS[2])
        self.plot_color_3 = Setting(start=DEFAULT_COLORS[3])
        self.plot_color_4 = Setting(start=DEFAULT_COLORS[4])

        # save changed settings to disk
        for _, setting in self:
            setting.add_callback(lambda _: save_settings(self), call_immediatly=False)

    def __iter__(self) -> Iterator[Tuple[str, Setting]]:
        for name, setting in self.__dict__.items():
            if isinstance(setting, Setting):
                yield name, setting


def save_settings(settings: Settings) -> None:
    data = {name: setting.value for name, setting in settings}
    with open(USER_DATA_PATH / SETTINGS_STORE_FILENAME, "w") as f:
        json.dump(data, f, indent=0)


def load_settings() -> Settings:
    settings = Settings()
    filename = USER_DATA_PATH / SETTINGS_STORE_FILENAME
    try:
        with open(filename, "r") as f:
            data = json.load(f)
            for name, value in data.items():
                if name in settings.__dict__:
                    getattr(settings, name).value = value
    except FileNotFoundError:
        save_settings(settings)
    except json.JSONDecodeError:
        logger.error(f"Settings file {filename} was corrupted.")
        create_backup_file(filename)
    return settings



# Copyright 2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QDoubleSpinBox, QSpinBox


class CustomDoubleSpinBox(QDoubleSpinBox):
    """
    Custom spin box with improved keyboard controls with step size depending on the
    cursor position.
    """

    def textFromValue(self, value):
        # show + sign for positive values
        text = super().textFromValue(value)
        if value >= 0:
            text = "+" + text
        return text

    def stepBy(self, steps):
        cursor_position = self.lineEdit().cursorPosition()
        # number of characters before the decimal separator including +/- sign
        n_chars_before_sep = len(str(abs(int(self.value())))) + 1
        if cursor_position == 0:
            # set the cursor right of the +/- sign
            self.lineEdit().setCursorPosition(1)
            cursor_position = self.lineEdit().cursorPosition()
        single_step = 10 ** (n_chars_before_sep - cursor_position)
        # Handle decimal separator. Step should be 0.1 if cursor is at `1.|23` or
        # `1.2|3`.
        if cursor_position >= n_chars_before_sep + 2:
            single_step = 10 * single_step
        # Change single step and perform the step
        self.setSingleStep(single_step)
        super().stepBy(steps)
        # Undo selection of the whole text.
        self.lineEdit().deselect()
        # Handle cases where the number of characters before the decimal separator
        # changes. Step size should remain the same.
        new_n_chars_before_sep = len(str(abs(int(self.value())))) + 1
        if new_n_chars_before_sep < n_chars_before_sep:
            cursor_position -= 1
        elif new_n_chars_before_sep > n_chars_before_sep:
            cursor_position += 1
        self.lineEdit().setCursorPosition(cursor_position)


class CustomDoubleSpinBoxNoSign(QDoubleSpinBox):
    """
    Custom spin box with improved keyboard controls with step size depending on the
    cursor position.
    """

    def stepBy(self, steps):
        cursor_position = self.lineEdit().cursorPosition()
        # number of characters before the decimal separator
        n_chars_before_sep = len(str(abs(int(self.value()))))
        n_chars_before_sep_max = len(str(abs(int(self.maximum()))))
        exponent = min(
            [n_chars_before_sep - cursor_position, n_chars_before_sep_max - 1]
        )
        single_step = 10**exponent
        # Handle decimal separator. Step should be 0.1 if cursor is at `1.|23` or
        # `1.2|3`.
        if cursor_position >= n_chars_before_sep + 2:
            single_step = 10 * single_step
        # Change single step and perform the step
        self.setSingleStep(single_step)
        super().stepBy(steps)
        # Undo selection of the whole text.
        self.lineEdit().deselect()
        # Handle cases where the number of characters before the decimal separator
        # changes. Step size should remain the same.
        new_n_chars_before_sep = len(str(abs(int(self.value()))))
        if new_n_chars_before_sep < n_chars_before_sep:
            cursor_position -= 1
        elif new_n_chars_before_sep > n_chars_before_sep:
            cursor_position += 1
        self.lineEdit().setCursorPosition(cursor_position)


class CustomSpinBox(QSpinBox):
    """
    Custom spin box with improved keyboard controls with step size depending on the
    cursor position. Works for positive integers only.
    """

    def stepBy(self, steps):
        n_chars = len(str(self.value()))  # number of characters of the current value
        n_chars_max = len(str(self.maximum()))  # number of characters of the maximum
        cursor_position = self.lineEdit().cursorPosition()
        exponent = min([n_chars - cursor_position, n_chars_max - 1])
        single_step = 10**exponent
        self.setSingleStep(single_step)
        super().stepBy(steps)
        # Undo selection of the whole text.
        self.lineEdit().deselect()
        # Handle cases where the number of characters before the decimal separator
        # changes. Step size should remain the same.
        new_n_chars = len(str(self.value()))
        if new_n_chars < n_chars:
            cursor_position -= 1
        elif new_n_chars > n_chars:
            cursor_position += 1
        self.lineEdit().setCursorPosition(cursor_position)



# Copyright 2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging

from linien_client.remote_parameters import RemoteParameters
from linien_common.influxdb import InfluxDBCredentials
from linien_gui.config import UI_PATH
from linien_gui.utils import get_linien_app_instance
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import pyqtSignal

START_LOG_BUTTON_TEXT = "Start Logging"
STOP_LOG_BUTTON_TEXT = "Stop Logging"

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class LoggingPanel(QtWidgets.QWidget):
    set_parameter_log = pyqtSignal(str, bool)
    influx_credentials_update = pyqtSignal(bool, str)

    logParametersToolButton: "LoggedParametersToolButton"
    lineEditURL: QtWidgets.QLineEdit
    lineEditOrg: QtWidgets.QLineEdit
    lineEditToken: QtWidgets.QLineEdit
    lineEditBucket: QtWidgets.QLineEdit
    lineEditMeas: QtWidgets.QLineEdit
    influxUpdateButton: QtWidgets.QPushButton
    influxTestIndicator: QtWidgets.QLabel
    logIntervalSpinBox: QtWidgets.QSpinBox
    logPushButton: QtWidgets.QPushButton

    def __init__(self, *args, **kwargs) -> None:
        super(LoggingPanel, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "logging_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.logParametersToolButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.loggedParametersMenu = LoggedParametersMenu()

    def on_connection_established(self) -> None:
        self.parameters = self.app.parameters
        self.control = self.app.control

        self.loggedParametersMenu.create_menu_entries(self.parameters)
        self.logParametersToolButton.setMenu(self.loggedParametersMenu)

        self.loggedParametersMenu.item_clicked.connect(
            self.on_parameter_log_status_changed
        )
        self.logPushButton.clicked.connect(self.on_logging_button_clicked)
        self.influxUpdateButton.clicked.connect(self.on_influx_update_button_clicked)
        self.influx_credentials_update.connect(self.on_influxdb_credentials_updated)

        # getting the influxdb credentials from the remote
        credentials = self.control.exposed_get_influxdb_credentials()
        logger.debug("Received InfluxDB credentials from server.")
        self.lineEditURL.setText(credentials.url)
        self.lineEditOrg.setText(credentials.org)
        self.lineEditToken.setText(credentials.token)
        self.lineEditBucket.setText(credentials.bucket)
        self.lineEditMeas.setText(credentials.measurement)

        # hide test indicator that might be set from previous connection to a different
        # device
        self.influxTestIndicator.setText("")

        # getting the logging status from the remote
        log_status = self.control.exposed_get_logging_status()
        button_text = STOP_LOG_BUTTON_TEXT if log_status else START_LOG_BUTTON_TEXT
        self.logPushButton.setChecked(log_status)
        self.logPushButton.setText(button_text)
        self.logIntervalSpinBox.setEnabled(not log_status)

    def on_parameter_log_status_changed(self, param_name: str, status: bool) -> None:
        self.control.exposed_set_parameter_log(param_name, status)

    def on_logging_button_clicked(self) -> None:
        if self.logPushButton.isChecked():
            self.control.exposed_start_logging(self.logIntervalSpinBox.value())
            self.logPushButton.setText(STOP_LOG_BUTTON_TEXT)
            self.logIntervalSpinBox.setEnabled(False)
        else:
            self.control.exposed_stop_logging()
            self.logPushButton.setText(START_LOG_BUTTON_TEXT)
            self.logIntervalSpinBox.setEnabled(True)

    def on_influx_update_button_clicked(self) -> None:
        credentials = InfluxDBCredentials(
            url=self.lineEditURL.text(),
            org=self.lineEditOrg.text(),
            token=self.lineEditToken.text(),
            bucket=self.lineEditBucket.text(),
            measurement=self.lineEditMeas.text(),
        )
        (success, message) = self.control.exposed_update_influxdb_credentials(
            credentials
        )
        update_msg = f"Update of InfluxDB credentials successful: {success}: {message}"
        logger.info(update_msg)
        self.influx_credentials_update.emit(success, message)

    def on_influxdb_credentials_updated(self, success: bool, message: str) -> None:
        if success:
            self.influxTestIndicator.setText("✅")
            self.influxTestIndicator.setToolTip(f"Connection successful: {message}")
        else:
            self.influxTestIndicator.setText("❌")
            self.influxTestIndicator.setToolTip(f"Connection failed: {message}.")


# checkable menu for logged parameters, inspired by
# https://stackoverflow.com/a/22775990/2750945
class LoggedParametersToolButton(QtWidgets.QToolButton):
    def __init__(self, *args, **kwargs):
        super(LoggedParametersToolButton, self).__init__(*args, **kwargs)


class LoggedParametersMenu(QtWidgets.QMenu):
    item_clicked = pyqtSignal(str, bool)

    def __init__(self, *args, **kwargs) -> None:
        super(LoggedParametersMenu, self).__init__(*args, **kwargs)

    def create_menu_entries(self, parameters: RemoteParameters) -> None:
        # remove all previous actions that might be present from a previous connection
        for action in self.actions():
            self.removeAction(action)
        param_dict = {name: param for name, param in parameters}
        param_dict = dict(sorted(param_dict.items()))  # sort alphabetically
        for name, param in param_dict.items():
            if param.loggable:
                action = QtWidgets.QAction(name, parent=self, checkable=True)  # type: ignore[call-overload] # noqa: E501
                action.setChecked(param.log)
                self.addAction(action)
        self.triggered.connect(self.on_item_selected)

    def on_item_selected(self, action: QtWidgets.QAction) -> None:
        param_name = action.text()
        self.item_clicked.emit(param_name, action.isChecked())



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import pickle
from time import time

import linien_gui
from linien_common.common import PSDAlgorithm
from linien_gui.config import UI_PATH
from linien_gui.dialogs import error_dialog
from linien_gui.utils import (
    RandomColorChoser,
    get_linien_app_instance,
    param2ui,
    set_window_icon,
)
from PyQt5 import QtWidgets, uic


class PSDWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super(PSDWindow, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "psd_window.ui", self)
        self.setWindowTitle("Linien: Noise analysis")
        set_window_icon(self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.random_color_choser = RandomColorChoser()
        self.colors = {}
        self.data = {}
        self.complete_uids = []

        self.start_psd_button.clicked.connect(self.start_psd)
        self.stop_psd_button.clicked.connect(self.stop_psd)

        self.curve_table.show_or_hide_curve.connect(
            self.psd_plot_widget.show_or_hide_curve
        )
        self.delete_curve_button.clicked.connect(self.delete_curve)
        self.export_psd_button.clicked.connect(self.export_psd)
        self.import_psd_button.clicked.connect(self.import_psd)

        self.maximum_measurement_time.currentIndexChanged.connect(
            self.change_maximum_measurement_time
        )

        self.psd_algorithm.currentIndexChanged.connect(self.change_psd_algorithm)

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        self.parameters.psd_data_partial.add_callback(
            self.psd_data_received,
        )
        self.parameters.psd_data_complete.add_callback(
            self.psd_data_received,
        )

        param2ui(
            self.parameters.psd_acquisition_max_decimation,
            self.maximum_measurement_time,
            lambda max_decimation: max_decimation - 12,
        )
        param2ui(self.parameters.psd_algorithm, self.psd_algorithm)

        def update_status(_):
            psd_running = self.parameters.psd_acquisition_running.value
            if psd_running:
                self.container_psd_running.show()
                self.container_psd_not_running.hide()
            else:
                self.container_psd_running.hide()
                self.container_psd_not_running.show()

        self.parameters.psd_acquisition_running.add_callback(update_status)

    def closeEvent(self, event, *args, **kwargs):
        # we never really want to close the window (which destroys its content)  but
        # just to hide it
        event.ignore()
        self.hide()

    def change_maximum_measurement_time(self, index):
        self.parameters.psd_acquisition_max_decimation.value = 12 + index

    def change_psd_algorithm(self, index):
        self.parameters.psd_algorithm.value = [PSDAlgorithm.LPSD, PSDAlgorithm.WELCH][
            index
        ]

    def psd_data_received(self, data_pickled):
        if data_pickled is None:
            return

        data = pickle.loads(data_pickled)

        curve_uuid = data["uuid"]

        if curve_uuid in self.complete_uids:
            # in networks with high latency it may happen that psd data is not received
            # in order. We do not want to update a complete plot with a partial one
            # --> stop here
            return

        if data["complete"]:
            self.complete_uids.append(curve_uuid)

        # either reuse the color of the previous partial plot of this curve or generate
        # a new color if the curve was not yet partially plotted
        if curve_uuid not in self.colors:
            color = self.random_color_choser.get()
            self.colors[curve_uuid] = color
        else:
            color = self.colors[curve_uuid]

        self.psd_plot_widget.plot_curve(curve_uuid, data["psds"], color)
        self.curve_table.add_curve(curve_uuid, data, color)

        self.data[curve_uuid] = data

    def start_psd(self):
        if not self.parameters.lock.value:
            return error_dialog(self, """Laser has to be locked for PSD measurement!""")

        self.control.start_psd_acquisition()

    def stop_psd(self):
        if self.parameters.task.value is not None:
            self.parameters.task.value.stop()
            self.parameters.task.value = None

    def start_pid_optimization(self):
        self.control.start_pid_optimization()

    def delete_curve(self):
        uuid = self.curve_table.delete_selected_curve()
        if uuid is not None:
            self.psd_plot_widget.delete_curve(uuid)
            del self.data[uuid]

    def export_psd(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".pickle"
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            f"PICKLE (*{default_ext})",
            options=options,
        )
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext

            with open(fn, "wb") as f:
                pickle.dump(
                    {
                        "linien-version": linien_gui.__version__,
                        "time": time(),
                        "psd-data": self.data,
                    },
                    f,
                )

    def import_psd(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".pickle"
        fn, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            f"JSON (*{default_ext})",
            options=options,
        )
        if fn:
            with open(fn, "rb") as f:
                data = pickle.load(f)

            assert "linien-version" in data, "invalid parameter file"

            for uuid, psd_data in data["psd-data"].items():
                self.psd_data_received(pickle.dumps(psd_data))



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import traceback
from typing import Dict, Tuple

from linien_client.connection import LinienClient
from linien_client.deploy import install_remote_server
from linien_client.device import Device, update_device
from linien_client.exceptions import (
    GeneralConnectionError,
    InvalidServerVersionException,
    RPYCAuthenticationException,
    ServerNotInstalledException,
)
from linien_client.remote_parameters import RemoteParameter
from linien_common.communication import RestorableParameterValues
from PyQt5.QtCore import QObject, QThread, pyqtSignal

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class RemoteOutStream(QObject):
    new_item = pyqtSignal(str)

    def write(self, item):
        self.new_item.emit(item)

    def read(self):
        pass

    def flush(self):
        pass


class RemoteServerInstallationThread(QThread):
    def __init__(self, device: Device) -> None:
        """A thread that installs the linien server on a remote machine."""
        super(RemoteServerInstallationThread, self).__init__()
        self.device = device

    out_stream = RemoteOutStream()

    def run(self) -> None:
        install_remote_server(self.device, out_stream=self.out_stream)


class ConnectionThread(QThread):
    def __init__(self, device: Device) -> None:
        super(ConnectionThread, self).__init__()
        self.device = device

    client_connected = pyqtSignal(object)
    server_not_installed_exception_raised = pyqtSignal()
    invalid_server_version_exception_raised = pyqtSignal(str, str)
    authentication_exception_raised = pyqtSignal()
    general_connection_exception_raised = pyqtSignal()
    other_exception_raised = pyqtSignal(str)
    connection_lost = pyqtSignal()
    parameter_difference = pyqtSignal(dict)

    def run(self) -> None:
        try:
            self.client = LinienClient(self.device)
            self.client.connect(
                autostart_server=True,
                use_parameter_cache=True,
                call_on_error=self.on_connection_lost,
            )
            self.client_connected.emit(self.client)

            # Check for locally cached settings for this server
            param_diff = self.compare_local_and_remote_parameters()
            if param_diff:
                self.parameter_difference.emit(param_diff)
            else:
                # if parameters don't differ, we can start monitoring remote parameter
                # changes and write them to disk. We don't do this if parameters differ
                # because we don't want to override our local settings with the remote
                # one --> we wait until user has answered whether local parameters or
                # remote ones should be used.
                self.add_callbacks_to_write_parameters_to_disk_on_change()

        except ServerNotInstalledException:
            self.server_not_installed_exception_raised.emit()

        except InvalidServerVersionException as e:
            self.invalid_server_version_exception_raised.emit(
                e.remote_version, e.client_version
            )

        except RPYCAuthenticationException:
            self.authentication_exception_raised.emit()

        except GeneralConnectionError:
            self.general_connection_exception_raised.emit()

        except Exception:
            traceback.print_exc()
            self.other_exception_raised.emit(traceback.format_exc())

    def on_connection_lost(self) -> None:
        self.connection_lost.emit()

    def compare_local_and_remote_parameters(
        self,
    ) -> Dict[str, Tuple[RestorableParameterValues, RestorableParameterValues]]:
        """Get differences between local and remote parameters."""
        differences = {}
        for local_param_name, local_param_value in self.device.parameters.items():
            if hasattr(self.client.parameters, local_param_name):
                remote_param: RemoteParameter = getattr(
                    self.client.parameters, local_param_name
                )
                if remote_param.value != local_param_value:
                    logger.info(
                        f"Parameter {local_param_name} differs: "
                        f"local={local_param_value}, remote={remote_param.value}"
                    )
                    differences[local_param_name] = (
                        local_param_value,
                        remote_param.value,
                    )
        return differences

    def restore_parameters(
        self,
        differences: Dict[
            str, Tuple[RestorableParameterValues, RestorableParameterValues]
        ],
    ) -> None:
        """Restore the remote parameters with the local ones."""
        logger.info("Restoring parameters...")
        for param_name, (local_value, remote_value) in differences.items():
            remote_param: RemoteParameter = getattr(self.client.parameters, param_name)
            remote_param.value = local_value
        self.client.control.exposed_write_registers()
        logger.info("Parameters restored.")

    def add_callbacks_to_write_parameters_to_disk_on_change(self) -> None:
        """
        Listens for changes of some parameters and permanently saves their values on the
        client's disk. This data can be used to restore the status later, if the client
        tries to connect to the server but it doesn't run anymore.
        """
        for param_name, param in self.client.parameters:
            if param.restorable:

                def on_change(value, parameter_name: str = param_name) -> None:
                    logger.debug(f"Parameter {parameter_name} changed to {value}")
                    if (
                        parameter_name not in self.device.parameters
                        or self.device.parameters[parameter_name] != value
                    ):
                        self.device.parameters[parameter_name] = value
                        update_device(self.device)

                param.add_callback(on_change)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import MHz, Vpp
from linien_gui.config import UI_PATH
from linien_gui.ui.spin_box import CustomDoubleSpinBoxNoSign
from linien_gui.utils import get_linien_app_instance, param2ui
from PyQt5 import QtWidgets, uic


class ModulationAndSweepPanel(QtWidgets.QWidget):
    sweepSpeedComboBox: QtWidgets.QGroupBox
    modulation_amplitude_group = QtWidgets.QGroupBox
    modulationAmplitudeSpinBox: CustomDoubleSpinBoxNoSign
    modulation_frequency_group = QtWidgets.QGroupBox
    modulationFrequencySpinBox: CustomDoubleSpinBoxNoSign
    spectroscopyTabs: QtWidgets.QTabWidget
    spectroscopy_channel_1_page: QtWidgets.QWidget
    spectroscopy_channel_2_page: QtWidgets.QWidget
    spectroscopyBPanelDisabled: QtWidgets.QLabel

    def __init__(self, *args, **kwargs):
        super(ModulationAndSweepPanel, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "modulation_sweep_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.modulationFrequencySpinBox.setKeyboardTracking(False)
        self.modulationFrequencySpinBox.valueChanged.connect(
            self.change_modulation_frequency
        )
        self.modulationAmplitudeSpinBox.setKeyboardTracking(False)
        self.modulationAmplitudeSpinBox.valueChanged.connect(
            self.change_modulation_amplitude
        )
        self.sweepSpeedComboBox.currentIndexChanged.connect(self.change_sweep_speed)

        self.spectroscopyTabs.setCurrentIndex(0)

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(
            self.parameters.modulation_frequency,
            self.modulationFrequencySpinBox,
            lambda value: value / MHz,
        )
        param2ui(
            self.parameters.modulation_amplitude,
            self.modulationAmplitudeSpinBox,
            lambda value: value / Vpp,
        )
        param2ui(self.parameters.sweep_speed, self.sweepSpeedComboBox)

        self.parameters.dual_channel.add_callback(self.dual_channel_changed)

        def pid_only_mode_changed(pid_only_mode_enabled):
            """Disables controls that are irrelevant if PID-only mode is enabled"""
            widgets_to_disable = (
                self.modulation_frequency_group,
                self.modulation_amplitude_group,
                self.spectroscopyTabs,
            )
            for widget in widgets_to_disable:
                widget.setEnabled(not pid_only_mode_enabled)

        self.parameters.pid_only_mode.add_callback(pid_only_mode_changed)

        def pid_only_mode_changed(pid_only_mode_enabled):
            """Disables controls that are irrelevant if PID-only mode is enabled"""
            widgets_to_disable = (
                self.modulation_frequency_group,
                self.modulation_amplitude_group,
                self.spectroscopyTabs,
            )
            for widget in widgets_to_disable:
                widget.setEnabled(not pid_only_mode_enabled)

        self.parameters.pid_only_mode.add_callback(pid_only_mode_changed)

    def change_modulation_frequency(self):
        self.parameters.modulation_frequency.value = (
            self.modulationFrequencySpinBox.value() * MHz
        )
        self.control.write_registers()

    def change_modulation_amplitude(self):
        self.parameters.modulation_amplitude.value = (
            self.modulationAmplitudeSpinBox.value() * Vpp
        )
        self.control.write_registers()

    def change_sweep_speed(self, decimation):
        self.parameters.sweep_speed.value = decimation
        self.control.write_registers()

    def dual_channel_changed(self, value):
        self.spectroscopyBPanel.setEnabled(value)
        self.spectroscopyBPanelDisabled.setVisible(not value)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from linien_gui.utils import color_to_hex, get_linien_app_instance
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import pyqtSignal


class PSDTableWidget(QtWidgets.QTableWidget):
    show_or_hide_curve = pyqtSignal(str, bool)

    def __init__(self, *args, **kwargs):
        super(PSDTableWidget, self).__init__(*args, **kwargs)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)

        self.uuids = []

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

    def add_curve(self, uuid, data, color):
        if uuid not in self.uuids:
            row_count = self.rowCount()
            self.setRowCount(row_count + 1)
            self.uuids.append(uuid)
        else:
            row_count = self.uuids.index(uuid)

        checkbox = QtWidgets.QCheckBox()

        checkbox.setChecked(True)
        checkbox.setStyleSheet("margin-left:auto; margin-right:auto;")
        checkbox.stateChanged.connect(
            lambda status, uuid=uuid: self.show_or_hide_curve.emit(uuid, status > 0)
        )
        self.setCellWidget(row_count, 0, checkbox)

        display_color = QtWidgets.QLabel()

        display_color.setStyleSheet("background-color: " + color_to_hex(color))
        self.setCellWidget(row_count, 1, display_color)

        def create_item(text):
            item = QtWidgets.QTableWidgetItem(str(text))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            return item

        time_formatted = datetime.utcfromtimestamp(data["time"]).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        self.setItem(row_count, 2, create_item(time_formatted))

        self.setItem(row_count, 3, create_item(data["p"]))
        self.setItem(row_count, 4, create_item(data["i"]))
        self.setItem(row_count, 5, create_item(data["d"]))
        self.setItem(row_count, 6, create_item(f"{data['fitness']:.4f}"))

        self.resizeColumnsToContents()

    def delete_selected_curve(self):
        current = self.currentRow()
        if not self.uuids or current == -1:
            return

        try:
            uuid = self.uuids.pop(current)
        except IndexError:
            return

        self.removeRow(current)

        # select the next row
        row_count = len(self.uuids)
        if row_count:
            if row_count > current:
                self.setCurrentCell(current, 0)
            else:
                self.setCurrentCell(current - 1, 0)

        return uuid



# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import superqt
from linien_gui.utils import get_linien_app_instance
from PyQt5 import QtCore, QtWidgets


class SweepControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(SweepControlWidget, self).__init__(*args, **kwargs)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)
        QtCore.QTimer.singleShot(100, self.ready)

    def ready(self):
        self.main_window = self.app.main_window
        self.main_window.sweepSlider.ready()  # initialize sweep slider boundaries

    def on_connection_established(self):
        self.control = self.app.control
        self.parameters = self.app.parameters

        self.main_window.sweepSlider.valueChanged.connect(self.update_sweep_range)
        # NOTE: The keyboardTracking property of the QDoubleSpinBoxes has been set to
        # False, to avoid signal emission when editing the field. Signals are still
        # emitted when using the arrow buttons. See also the editingFinished method.
        self.main_window.sweepCenterSpinBox.valueChanged.connect(
            self.update_sweep_center
        )
        self.main_window.sweepAmplitudeSpinBox.valueChanged.connect(
            self.update_sweep_amplitude
        )
        self.main_window.sweepStartStopPushButton.clicked.connect(
            self.pause_or_resume_sweep
        )

        # initialize sweep controls
        self.display_sweep_status()

        # change displayed values when sweep parameters change
        self.parameters.sweep_center.add_callback(self.display_sweep_status)
        self.parameters.sweep_amplitude.add_callback(self.display_sweep_status)
        self.parameters.sweep_pause.add_callback(self.display_sweep_status)

    def display_sweep_status(self, *args):
        center = self.parameters.sweep_center.value
        amplitude = self.parameters.sweep_amplitude.value
        min_ = center - amplitude
        max_ = center + amplitude

        # block signals to avoid infinite loops when changing sweep parameters, see also
        # param2ui
        self.main_window.sweepSlider.blockSignals(True)
        self.main_window.sweepAmplitudeSpinBox.blockSignals(True)
        self.main_window.sweepCenterSpinBox.blockSignals(True)

        self.main_window.sweepSlider.setValue((min_, max_))
        self.main_window.sweepCenterSpinBox.setValue(center)
        self.main_window.sweepAmplitudeSpinBox.setValue(amplitude)
        if self.parameters.sweep_pause.value:
            self.main_window.sweepStartStopPushButton.setText("Start")
        else:
            self.main_window.sweepStartStopPushButton.setText("Pause")

        self.main_window.sweepSlider.blockSignals(False)
        self.main_window.sweepCenterSpinBox.blockSignals(False)
        self.main_window.sweepAmplitudeSpinBox.blockSignals(False)

    def pause_or_resume_sweep(self):
        # If sweep is paused, resume it or vice versa.
        self.parameters.sweep_pause.value = not self.parameters.sweep_pause.value
        self.control.write_registers()

    def update_sweep_center(self, center):
        self.parameters.sweep_center.value = center
        self.control.write_registers()

    def update_sweep_amplitude(self, amplitude):
        self.parameters.sweep_amplitude.value = amplitude
        self.control.write_registers()

    def update_sweep_range(self, range_):
        min_, max_ = range_
        amplitude = (max_ - min_) / 2
        center = (max_ + min_) / 2
        self.parameters.sweep_amplitude.value = amplitude
        self.parameters.sweep_center.value = center
        self.control.write_registers()


class SweepSlider(superqt.QDoubleRangeSlider):
    def __init__(self, *args, **kwargs):
        super(SweepSlider, self).__init__(*args, **kwargs)

    def ready(self):
        # set control boundaries
        self.setMinimum(-1.0)
        self.setMaximum(1.0)
        self.setSingleStep(0.001)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging

import linien_gui
import requests
from packaging import version
from PyQt5.QtCore import QThread, pyqtSignal

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class VersionCheckerThread(QThread):
    check_done = pyqtSignal(bool)

    def run(self):
        our_version = version.parse(linien_gui.__version__)
        latest_version = our_version
        logger.debug("Check whether new version is available.")
        response = requests.get("https://pypi.org/pypi/linien-gui/json")
        if response.ok:
            latest_version = version.parse(response.json()["info"]["version"])
            self.check_done.emit(latest_version > our_version)
        else:
            logger.error("Failed to check for new version.")
            self.check_done.emit(False)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pyqtgraph as pg
from linien_gui.ui.plot_widget import V
from linien_gui.utils import get_linien_app_instance


class CustomLogAxis(pg.AxisItem):
    """This class overrides the tick label generator function of the default
    axis. It reimplements it such that negative exponents are also displayed
    in scientific notation."""

    def __init__(self, *args, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)

    def logTickStrings(self, values, scale, spacing):
        # this method is mainly taken from pyqtgraph, just taking care that negative
        # exponents are also displayed in scientific notation
        estrings = [
            f"{x:0.1e}" for x in 10 ** np.array(values).astype(float) * np.array(scale)
        ]

        convdict = {
            "0": "⁰",
            "1": "¹",
            "2": "²",
            "3": "³",
            "4": "⁴",
            "5": "⁵",
            "6": "⁶",
            "7": "⁷",
            "8": "⁸",
            "9": "⁹",
        }
        dstrings = []
        for e in estrings:
            if e.count("e"):
                v, p = e.split("e")

                sign = "⁻" if p[0] == "-" else ""
                if p[1:].count("0") == len(p[1:]):
                    pot = convdict["0"]
                else:
                    pot = "".join([convdict[pp] for pp in p[1:].lstrip("0")])

                v = v.rstrip(".0")

                if v == "1":
                    v = ""
                else:
                    v = v + "·"
                dstrings.append(v + "10" + sign + pot)
            else:
                dstrings.append(e)

        return dstrings


class PSDPlotWidget(pg.PlotWidget):
    def __init__(self, *args, **kwargs):
        super(PSDPlotWidget, self).__init__(
            *args,
            axisItems={
                "bottom": CustomLogAxis(orientation="bottom"),
                "left": CustomLogAxis(orientation="left"),
            },
            **kwargs,
        )
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.curves = {}

        self.setLogMode(x=True, y=True)
        self.setLabel("left", "PSD", units="V / Sqrt[Hz]")
        self.setLabel("bottom", "Frequency", units="Hz")
        self.getAxis("left").enableAutoSIPrefix(False)
        self.getAxis("bottom").enableAutoSIPrefix(False)
        self.showGrid(x=True, y=True)

        self.vertical_line = pg.InfiniteLine(angle=90, movable=False)
        self.vertical_line.hide()
        self.horizontal_line = pg.InfiniteLine(angle=0, movable=False)
        self.horizontal_line.hide()
        self.cursor_label = pg.TextItem()
        self.scene().sigMouseMoved.connect(self.mouseMoved)

        for element in (self.vertical_line, self.horizontal_line, self.cursor_label):
            # ignoreBounds tells pyqtgraph not to consider these items when calculating
            # automatic axis scaling
            self.addItem(element, ignoreBounds=True)
            # all these elements should be on top of all plots
            element.setZValue(10000)

        self.recalculate_min_max()

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

    def plot_curve(self, uuid, psds, color):
        self.curves[uuid] = self.curves.get(uuid, [])
        for idx in range(len(psds)):
            if len(self.curves[uuid]) <= idx:
                curve = pg.PlotCurveItem()
                self.curves[uuid].append(curve)
                self.addItem(curve)

        # sort such that high decimations are first
        psds_sorted = sorted(psds.items(), key=lambda v: -1 * v[0])
        highest_plotted_frequency = 0
        highest_plotted_frequency_psd = None

        for idx, [decimation, [f, psd]] in enumerate(psds_sorted):
            curve = self.curves[uuid][idx]

            psd = psd[f > highest_plotted_frequency]
            f = f[f > highest_plotted_frequency]

            if highest_plotted_frequency_psd is not None:
                # this connects first and last points of adjacent segments
                f = np.array([highest_plotted_frequency] + list(f))
                psd = np.array([highest_plotted_frequency_psd] + list(psd))

            highest_plotted_frequency = f[-1]
            highest_plotted_frequency_psd = psd[-1]

            curve.setData(np.log10(f), np.log10(psd / V))
            r, g, b = color
            curve.setPen(pg.mkPen((r, g, b, 200)))

        self.recalculate_min_max()

    def recalculate_min_max(self):
        self._x_min = np.inf
        self._x_max = -np.inf
        self._y_min = np.inf
        self._y_max = -np.inf

        if self.curves:
            for curves in self.curves.values():
                for curve in curves:
                    x, y = curve.getData()
                    if np.min(x) < self._x_min:
                        self._x_min = np.min(x)
                    if np.max(x) > self._x_max:
                        self._x_max = np.max(x)
                    if np.min(y) < self._y_min:
                        self._y_min = np.min(y)
                    if np.max(y) > self._y_max:
                        self._y_max = np.max(y)

    def show_or_hide_curve(self, uuid, show):
        curves = self.curves.get(uuid, [])

        for curve in curves:
            curve.setVisible(show)

    def delete_curve(self, uuid):
        for curve in self.curves[uuid]:
            self.removeItem(curve)
        del self.curves[uuid]

    def _to_data_coords(self, event):
        pos = self.plotItem.vb.mapSceneToView(event.pos())
        x, y = pos.x(), pos.y()
        return x, y

    def mouseMoved(self, evt):
        pos = evt

        if self.sceneBoundingRect().contains(pos):
            mousePoint = self.plotItem.vb.mapSceneToView(pos)
            x = mousePoint.x()
            y = mousePoint.y()

            # if index > 0 and index < self.MFmax:
            self.cursor_label.setHtml(
                f"<span style='font-size: 12pt'>({10**x:.1e},{10**y:.1e})</span>"
            )
            # this determines whether cursor label is on right or left side of
            # crosshair
            self.cursor_label.setAnchor((1 if x > self._x_max / 2 else 0, 1))
            self.cursor_label.setPos(x, y)
            self.vertical_line.setPos(x)
            self.horizontal_line.setPos(y)

            self.show_cursor_position(True)
        else:
            self.show_cursor_position(False)

    def show_cursor_position(self, show):
        self.vertical_line.setVisible(show)
        self.horizontal_line.setVisible(show)
        self.cursor_label.setVisible(show)

    def leaveEvent(self, QEvent):
        """This method is called when the mouse used to hover the plot and
        not left. In this case, we hide crosshair and cursor label."""
        super().leaveEvent(QEvent)
        self.show_cursor_position(False)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import AutolockMode
from linien_gui.config import UI_PATH
from linien_gui.ui.lock_status_panel import LockStatusPanel
from linien_gui.ui.spin_box import CustomSpinBox
from linien_gui.utils import get_linien_app_instance, param2ui
from PyQt5 import QtWidgets, uic


class LockingPanel(QtWidgets.QWidget):
    kpSpinBox: CustomSpinBox
    kiSpinBox: CustomSpinBox
    kdSpinBox: CustomSpinBox
    slow_pid_group: QtWidgets.QGroupBox
    pid_on_slow_strength: CustomSpinBox
    lock_control_container: QtWidgets.QTabWidget
    auto_mode_activated: QtWidgets.QWidget
    abortLineSelection: QtWidgets.QPushButton
    auto_mode_not_activated: QtWidgets.QWidget
    autoOffsetCheckbox: QtWidgets.QCheckBox
    autolock_mode_preference: QtWidgets.QComboBox
    selectLineToLock: QtWidgets.QPushButton
    manual_mode: QtWidgets.QWidget
    button_slope_falling: QtWidgets.QRadioButton
    button_slope_rising: QtWidgets.QRadioButton
    manualLockButton: QtWidgets.QPushButton
    lock_failed: QtWidgets.QWidget
    reset_lock_failed_state: QtWidgets.QPushButton
    lock_status_container: LockStatusPanel
    controlSignalHistoryLengthSpinBox: CustomSpinBox
    lock_status: QtWidgets.QLabel
    stopLockPushButton: QtWidgets.QPushButton

    def __init__(self, *args, **kwargs):
        super(LockingPanel, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "locking_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.kpSpinBox.valueChanged.connect(self.kp_changed)
        self.kiSpinBox.valueChanged.connect(self.ki_changed)
        self.kdSpinBox.valueChanged.connect(self.kd_changed)
        self.lock_control_container.currentChanged.connect(self.lock_mode_changed)

        self.selectLineToLock.clicked.connect(self.start_autolock_selection)
        self.abortLineSelection.clicked.connect(self.stop_autolock_selection)

        self.manualLockButton.clicked.connect(self.start_manual_lock)
        self.autoOffsetCheckbox.stateChanged.connect(self.auto_offset_changed)

        self.pid_on_slow_strength.setKeyboardTracking(False)
        self.pid_on_slow_strength.valueChanged.connect(
            self.pid_on_slow_strength_changed
        )

        self.reset_lock_failed_state.clicked.connect(self.reset_lock_failed)

        self.autolock_mode_preference.currentIndexChanged.connect(
            self.autolock_mode_preference_changed
        )

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.parameters.p, self.kpSpinBox)
        param2ui(self.parameters.i, self.kiSpinBox)
        param2ui(self.parameters.d, self.kdSpinBox)

        param2ui(self.parameters.autolock_determine_offset, self.autoOffsetCheckbox)
        param2ui(
            self.parameters.automatic_mode,
            self.lock_control_container,
            lambda value: 0 if value else 1,
        )
        param2ui(self.parameters.pid_on_slow_strength, self.pid_on_slow_strength)

        def slow_pid_visibility(*args):
            self.slow_pid_group.setVisible(self.parameters.pid_on_slow_enabled.value)

        self.parameters.pid_on_slow_enabled.add_callback(slow_pid_visibility)

        def lock_status_changed(_):
            locked = self.parameters.lock.value
            task = self.parameters.task.value
            al_failed = self.parameters.autolock_failed.value
            task_running = (task is not None) and (not al_failed)

            if locked or task_running or al_failed:
                self.lock_control_container.hide()
            else:
                self.lock_control_container.show()

            self.lock_failed.setVisible(al_failed)

        for param in (
            self.parameters.lock,
            self.parameters.autolock_preparing,
            self.parameters.autolock_watching,
            self.parameters.autolock_failed,
            self.parameters.autolock_locked,
        ):
            param.add_callback(lock_status_changed)

        param2ui(self.parameters.target_slope_rising, self.button_slope_rising)
        param2ui(
            self.parameters.target_slope_rising,
            self.button_slope_falling,
            lambda value: not value,
        )

        def autolock_selection_status_changed(value):
            self.auto_mode_activated.setVisible(value)
            self.auto_mode_not_activated.setVisible(not value)

        self.parameters.autolock_selection.add_callback(
            autolock_selection_status_changed
        )

        param2ui(
            self.parameters.autolock_mode_preference, self.autolock_mode_preference
        )

    def kp_changed(self):
        self.parameters.p.value = self.kpSpinBox.value()
        self.control.write_registers()

    def ki_changed(self):
        self.parameters.i.value = self.kiSpinBox.value()
        self.control.write_registers()

    def kd_changed(self):
        self.parameters.d.value = self.kdSpinBox.value()
        self.control.write_registers()

    def lock_mode_changed(self, idx):
        self.parameters.automatic_mode.value = idx == 0

    def autolock_mode_preference_changed(self, idx):
        self.parameters.autolock_mode_preference.value = idx

    def start_manual_lock(self):
        self.parameters.target_slope_rising.value = self.button_slope_rising.isChecked()
        self.parameters.fetch_additional_signals.value = False
        self.parameters.autolock_mode.value = AutolockMode.SIMPLE
        self.parameters.autolock_target_position.value = 0
        self.control.write_registers()
        self.control.start_lock()

    def auto_offset_changed(self):
        self.parameters.autolock_determine_offset.value = int(
            self.autoOffsetCheckbox.checkState()
        )

    def pid_on_slow_strength_changed(self):
        self.parameters.pid_on_slow_strength.value = self.pid_on_slow_strength.value()
        self.control.write_registers()

    def start_autolock_selection(self):
        self.parameters.autolock_selection.value = True

    def stop_autolock_selection(self):
        self.parameters.autolock_selection.value = False

    def reset_lock_failed(self):
        self.parameters.autolock_failed.value = False



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable

from linien_client.device import Device
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (
    QDialog,
    QListWidget,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from pyqtgraph import QtCore

from .threads import RemoteServerInstallationThread


class SSHCommandOutputWidget(QListWidget):
    command_finished = pyqtSignal()
    enable_button = pyqtSignal()

    def __init__(self, parent: QWidget):
        super(SSHCommandOutputWidget, self).__init__(parent)
        self.setSelectionMode(self.NoSelection)

    def run(self, thread):
        self.thread = thread
        self.thread.out_stream.new_item.connect(self.on_new_item_in_out_stream)
        self.thread.finished.connect(self.on_thread_finished)
        self.thread.start()

    def on_new_item_in_out_stream(self, line: str):
        self.addItem(line.rstrip())
        self.scrollToBottom()

    def on_thread_finished(self):
        self.addItem("\nFinished.")
        self.scrollToBottom()
        self.command_finished.emit()


def show_installation_progress_widget(
    parent: QWidget, device: Device, callback: Callable
):
    window = QDialog(parent)
    window.setWindowTitle("Deploying Linien Server")
    window.resize(800, 600)
    window_layout = QVBoxLayout(window)
    widget = SSHCommandOutputWidget(parent)
    button = QPushButton("Continue")
    button.setEnabled(False)
    window_layout.addWidget(widget)
    window_layout.addWidget(button)
    window.setLayout(window_layout)
    window.setModal(True)
    window.setWindowModality(QtCore.Qt.WindowModal)
    window.show()

    widget.command_finished.connect(lambda: button.setEnabled(True))
    button.clicked.connect(callback)
    button.clicked.connect(window.close)

    thread = RemoteServerInstallationThread(device)
    widget.run(thread)

    return window


class LoadingDialog(QMessageBox):
    aborted = pyqtSignal()

    def __init__(self, parent: QWidget, host: str):
        super(LoadingDialog, self).__init__(parent)

        self.setIcon(QMessageBox.Information)
        self.setText(f"Connecting to {host}")
        self.setWindowTitle("Connecting")
        self.setModal(True)
        self.setWindowModality(QtCore.Qt.WindowModal)
        self.setStandardButtons(QMessageBox.NoButton)
        self.show()

    def closeEvent(self, *args):
        self.aborted.emit()

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Escape:
            self.close()


def error_dialog(parent: QWidget, error):
    return QMessageBox.question(parent, "Error", error, QMessageBox.Ok, QMessageBox.Ok)


def question_dialog(parent, question, title):
    box = QMessageBox(parent)
    box.setText(question)
    box.setWindowTitle(title)
    reply = QMessageBox.question(
        parent, title, question, QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes
    )

    return reply == QMessageBox.Yes


def ask_for_parameter_restore_dialog(parent, question, title):
    box = QMessageBox(parent)
    box.setText(question)
    box.setWindowTitle(title)
    # do_nothing_button
    _ = box.addButton("Keep remote parameters", QMessageBox.NoRole)
    upload_button = box.addButton("Upload local parameters", QMessageBox.YesRole)

    box.exec_()

    return box.clickedButton() == upload_button



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
# Copyright 2022 Christian Freier <christian.freier@nomadatomics.com>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import (
    ANALOG_OUT_V,
    OutputChannel,
    convert_channel_mixing_value,
)
from linien_gui.config import UI_PATH
from linien_gui.ui.spin_box import CustomDoubleSpinBoxNoSign
from linien_gui.utils import get_linien_app_instance, param2ui
from PyQt5 import QtWidgets, uic


class GeneralPanel(QtWidgets.QWidget):
    analogOutComboBox1: CustomDoubleSpinBoxNoSign
    analogOutComboBox2: CustomDoubleSpinBoxNoSign
    analogOutComboBox3: CustomDoubleSpinBoxNoSign
    pidOnlyModeCheckBox: QtWidgets.QCheckBox
    fast_in_1_status: QtWidgets.QLabel
    fast_in_2_status: QtWidgets.QLabel
    dualChannelCheckBox: QtWidgets.QCheckBox
    dualChannelMixingGroupBox: QtWidgets.QGroupBox
    channelMixingSlider: QtWidgets.QSlider
    output_ports_group: QtWidgets.QGroupBox
    controlChannelComboBox: QtWidgets.QComboBox
    sweepChannelComboBox: QtWidgets.QComboBox
    slowControlComboBox: QtWidgets.QComboBox
    polaritySelectorGroupBox: QtWidgets.QGroupBox
    polarityContainerAnalogOut0: QtWidgets.QWidget
    polarityComboBoxAnalogOut0: QtWidgets.QComboBox
    polarityContainerFastOut1: QtWidgets.QWidget
    polarityComboBoxFastOut1: QtWidgets.QComboBox
    polarityContainerFastOut2: QtWidgets.QWidget
    polarityComboBoxFastOut2: QtWidgets.QComboBox
    modulationChannelComboBox: QtWidgets.QComboBox

    def __init__(self, *args, **kwargs) -> None:
        super(GeneralPanel, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "general_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.channelMixingSlider.valueChanged.connect(self.on_channel_mixing_changed)
        self.pidOnlyModeCheckBox.stateChanged.connect(self.on_pid_only_mode_changed)
        self.dualChannelCheckBox.stateChanged.connect(self.on_dual_channel_changed)
        self.modulationChannelComboBox.currentIndexChanged.connect(
            self.on_mod_channel_changed
        )
        self.controlChannelComboBox.currentIndexChanged.connect(
            self.on_control_channel_changed
        )
        self.sweepChannelComboBox.currentIndexChanged.connect(
            self.on_sweep_channel_changed
        )
        self.slowControlComboBox.currentIndexChanged.connect(
            self.on_slow_control_channel_changed
        )
        self.polarityComboBoxFastOut1.currentIndexChanged.connect(
            self.on_polarity_fast_out1_changed
        )
        self.polarityComboBoxFastOut2.currentIndexChanged.connect(
            self.on_polarity_fast_out2_changed
        )
        self.polarityComboBoxAnalogOut0.currentIndexChanged.connect(
            self.on_polarity_analog_out0_changed
        )

        for idx in range(1, 4):
            element: CustomDoubleSpinBoxNoSign = getattr(
                self, f"analogOutComboBox{idx}"
            )
            element.setKeyboardTracking(False)
            element.valueChanged.connect(
                lambda _, idx=idx: self.on_analog_out_changed(idx)
            )

    def on_connection_established(self) -> None:
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.parameters.pid_only_mode, self.pidOnlyModeCheckBox)

        def on_dual_channel_changed(value):
            self.dualChannelMixingGroupBox.setVisible(value)
            self.fast_in_1_status.setText(
                "error signal" if not value else "error signal 1"
            )
            self.fast_in_2_status.setText("monitor" if not value else "error signal 2")
            return value

        param2ui(
            self.parameters.dual_channel,
            self.dualChannelCheckBox,
            on_dual_channel_changed,
        )

        param2ui(
            self.parameters.channel_mixing,
            self.channelMixingSlider,
            lambda value: value + 128,
        )
        # this is required to update the descriptive labels in the beginning
        self.on_channel_mixing_changed()

        param2ui(self.parameters.mod_channel, self.modulationChannelComboBox)
        param2ui(self.parameters.control_channel, self.controlChannelComboBox)
        param2ui(self.parameters.sweep_channel, self.sweepChannelComboBox)
        param2ui(self.parameters.slow_control_channel, self.slowControlComboBox)

        param2ui(self.parameters.polarity_fast_out1, self.polarityComboBoxFastOut1)
        param2ui(self.parameters.polarity_fast_out2, self.polarityComboBoxFastOut2)
        param2ui(self.parameters.polarity_analog_out0, self.polarityComboBoxAnalogOut0)

        self.parameters.control_channel.add_callback(self.show_polarity_settings)
        self.parameters.sweep_channel.add_callback(self.show_polarity_settings)
        self.parameters.mod_channel.add_callback(self.show_polarity_settings)
        self.parameters.slow_control_channel.add_callback(self.show_polarity_settings)
        self.parameters.pid_on_slow_enabled.add_callback(self.show_polarity_settings)

        for idx in range(1, 4):
            param2ui(
                getattr(self.parameters, f"analog_out_{idx}"),
                getattr(self, f"analogOutComboBox{idx}"),
                process_value=lambda v: ANALOG_OUT_V * v,
            )

        def on_pid_only_mode_changed(pid_only_mode_enabled):
            """Disables controls that are irrelevant if PID-only mode is enabled"""
            widgets_to_disable = (
                self.output_ports_group,
                self.input_ports_group,
            )
            for widget in widgets_to_disable:
                widget.setEnabled(not pid_only_mode_enabled)

        self.parameters.pid_only_mode.add_callback(on_pid_only_mode_changed)

    def on_analog_out_changed(self, idx):
        getattr(self.parameters, f"analog_out_{idx}").value = int(
            getattr(self, f"analogOutComboBox{idx}").value() / ANALOG_OUT_V
        )
        self.control.write_registers()

    def on_channel_mixing_changed(self):
        value = int(self.channelMixingSlider.value()) - 128
        self.parameters.channel_mixing.value = value
        self.control.write_registers()

        # update channel mixing slider
        a_value, b_value = convert_channel_mixing_value(value)
        self.chain_a_factor.setText(f"{a_value}")
        self.chain_b_factor.setText(f"{b_value}")

    def on_pid_only_mode_changed(self):
        self.parameters.pid_only_mode.value = int(
            self.pidOnlyModeCheckBox.checkState() > 0
        )
        self.control.write_registers()

    def on_dual_channel_changed(self):
        self.parameters.dual_channel.value = int(
            self.dualChannelCheckBox.checkState() > 0
        )
        self.control.write_registers()

    def on_mod_channel_changed(self, channel):
        self.parameters.mod_channel.value = channel
        self.control.write_registers()

    def on_control_channel_changed(self, channel):
        self.parameters.control_channel.value = channel
        self.control.write_registers()

    def on_slow_control_channel_changed(self, channel):
        if channel > 2:
            # disabled state
            self.parameters.pid_on_slow_enabled.value = False
        else:
            self.parameters.slow_control_channel.value = channel
            self.parameters.pid_on_slow_enabled.value = True
        self.control.write_registers()

    def on_sweep_channel_changed(self, channel):
        self.parameters.sweep_channel.value = channel
        self.control.write_registers()

    def on_polarity_fast_out1_changed(self, polarity):
        self.parameters.polarity_fast_out1.value = bool(polarity)
        self.control.write_registers()

    def on_polarity_fast_out2_changed(self, polarity):
        self.parameters.polarity_fast_out2.value = bool(polarity)
        self.control.write_registers()

    def on_polarity_analog_out0_changed(self, polarity):
        self.parameters.polarity_analog_out0.value = bool(polarity)
        self.control.write_registers()

    def show_polarity_settings(self, *args):
        used_channels = {
            self.parameters.control_channel.value,
            self.parameters.sweep_channel.value,
        }

        if self.parameters.pid_on_slow_enabled.value:
            used_channels.add(self.parameters.slow_control_channel.value)

        self.polaritySelectorGroupBox.setVisible(len(used_channels) > 1)
        self.polarityContainerFastOut1.setVisible(
            OutputChannel.FAST_OUT1 in used_channels
        )
        self.polarityContainerFastOut2.setVisible(
            OutputChannel.FAST_OUT2 in used_channels
        )
        self.polarityContainerAnalogOut0.setVisible(
            OutputChannel.ANALOG_OUT0 in used_channels
        )



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import signal
import sys

import click
from linien_client.connection import LinienClient
from linien_gui import __version__
from linien_gui.config import UI_PATH, load_settings
from linien_gui.ui.device_manager import DeviceManager
from linien_gui.ui.main_window import MainWindow
from linien_gui.ui.psd_window import PSDWindow
from linien_gui.ui.version_checker import VersionCheckerThread
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal
from pyqtgraph.Qt import QtCore

sys.path += [str(UI_PATH)]

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class LinienApp(QtWidgets.QApplication):
    connection_established = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(LinienApp, self).__init__(*args, **kwargs)

        self.settings = load_settings()

        self.main_window = MainWindow()
        self.device_manager = DeviceManager()
        self.psd_window = PSDWindow()
        self.device_manager.show()

        self.aboutToQuit.connect(self.quit)

    def client_connected(self, client: LinienClient):
        self.device_manager.hide()
        self.main_window.show(client.device.host, client.device.name)

        self.client = client
        self.control = client.control
        self.parameters = client.parameters

        self.connection_established.emit()

        self.periodically_check_for_changed_parameters()

        self.check_for_new_version()

    def periodically_check_for_changed_parameters(self):
        if hasattr(self, "client") and self.client and self.client.connected:
            try:
                self.parameters.check_for_changed_parameters()
            except AttributeError:
                logger.exception("check_for_changed_parameters() failed")

            QtCore.QTimer.singleShot(50, self.periodically_check_for_changed_parameters)

    def shutdown(self):
        self.client.control.exposed_shutdown()
        self.quit()

    def open_psd_window(self):
        # first hiding it, then showing it brings it to foregroud if it is in background
        self.psd_window.hide()
        self.psd_window.show()

    def open_device_manager(self):
        self.main_window.hide()
        self.device_manager.show()

        self.client.disconnect()
        del self.client

    def close_all_secondary_windows(self):
        self.psd_window.hide()

    def check_for_new_version(self):
        self.version_checker = VersionCheckerThread()
        self.version_checker.check_done.connect(self.new_version_available)
        self.version_checker.start()

    def new_version_available(self, new_version_available):
        if new_version_available:
            logger.info("New version available")
            self.main_window.show_new_version_available()
        else:
            logger.info("No new version available")
            QtCore.QTimer.singleShot(1000 * 60 * 60, self.check_for_new_version)


# ignore type, otherwise "Argument 1 has incompatible type "Callable[[int, bool, str |
# None, bool], Any]"; expected <nothing>" is raised for click 8.1.4.
@click.command("linien")  # type: ignore[arg-type]
@click.version_option(__version__)
def main():
    app = LinienApp(sys.argv)
    logger.info("Starting Linien GUI")

    # catch ctrl-c and shutdown
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import pickle
from os import path

import numpy as np
from linien_gui.config import N_COLORS, UI_PATH
from linien_gui.ui.spin_box import CustomDoubleSpinBoxNoSign, CustomSpinBox
from linien_gui.utils import color_to_hex, get_linien_app_instance, param2ui
from PyQt5 import QtGui, QtWidgets, uic


class ViewPanel(QtWidgets.QWidget):
    plot_line_width: CustomDoubleSpinBoxNoSign
    plot_line_opacity: CustomSpinBox
    plot_fill_opacity: CustomSpinBox
    display_color_0: QtWidgets.QLabel
    display_color_1: QtWidgets.QLabel
    display_color_2: QtWidgets.QLabel
    display_color_3: QtWidgets.QLabel
    display_color_4: QtWidgets.QLabel
    edit_color_0: QtWidgets.QToolButton
    edit_color_1: QtWidgets.QToolButton
    edit_color_2: QtWidgets.QToolButton
    edit_color_3: QtWidgets.QToolButton
    edit_color_4: QtWidgets.QToolButton
    export_data: QtWidgets.QPushButton
    export_select_file: QtWidgets.QPushButton

    def __init__(self, *args, **kwargs):
        super(ViewPanel, self).__init__(*args, **kwargs)
        uic.loadUi(UI_PATH / "view_panel.ui", self)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)

        self.export_select_file.clicked.connect(self.do_export_select_file)
        self.export_data.clicked.connect(self.do_export_data)

        self.plot_line_width.setKeyboardTracking(False)
        self.plot_line_width.valueChanged.connect(self.plot_line_width_changed)

        self.plot_line_opacity.setKeyboardTracking(False)
        self.plot_line_opacity.valueChanged.connect(self.plot_line_opacity_changed)

        self.plot_fill_opacity.setKeyboardTracking(False)
        self.plot_fill_opacity.valueChanged.connect(self.plot_fill_opacity_changed)

        for color_idx in range(N_COLORS):
            getattr(self, f"edit_color_{color_idx}").clicked.connect(
                lambda *args, color_idx=color_idx: self.edit_color(color_idx)
            )

    def edit_color(self, color_idx):
        setting = getattr(self.app.settings, f"plot_color_{color_idx}")
        color = QtWidgets.QColorDialog.getColor(QtGui.QColor.fromRgb(*setting.value))
        r, g, b, a = color.getRgb()
        getattr(self.app.settings, f"plot_color_{color_idx}").value = (r, g, b, a)

    def on_connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.app.settings.plot_line_width, self.plot_line_width)
        param2ui(self.app.settings.plot_line_opacity, self.plot_line_opacity)
        param2ui(self.app.settings.plot_fill_opacity, self.plot_fill_opacity)

        def preview_colors(*args):
            for color_idx in range(N_COLORS):
                element = getattr(self, f"display_color_{color_idx}")
                setting = getattr(self.app.settings, f"plot_color_{color_idx}")
                element.setStyleSheet(
                    f"background-color: {color_to_hex(setting.value)}"
                )

        for color_idx in range(N_COLORS):
            getattr(self.app.settings, f"plot_color_{color_idx}").add_callback(
                preview_colors
            )

    def plot_line_width_changed(self):
        self.app.settings.plot_line_width.value = self.plot_line_width.value()

    def plot_line_opacity_changed(self):
        self.app.settings.plot_line_opacity.value = self.plot_line_opacity.value()

    def plot_fill_opacity_changed(self):
        self.app.settings.plot_fill_opacity.value = self.plot_fill_opacity.value()

    def do_export_select_file(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".json"
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            f"JSON (*{default_ext})",
            options=options,
        )
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext
            self.export_fn = fn
            self.export_select_file.setText(f"File selected: {path.split(fn)[-1]}")
            self.export_data.setEnabled(True)

    def do_export_data(self):
        fn = self.export_fn
        counter = 0

        while True:
            if counter > 0:
                name, ext = path.splitext(fn)
                fn_with_suffix = name + "-" + str(counter)
                if ext:
                    fn_with_suffix += ext
            else:
                fn_with_suffix = fn

            try:
                open(fn_with_suffix, "r")
                counter += 1
                continue
            except FileNotFoundError:
                break

        print(f"export data to {fn_with_suffix}")

        with open(fn_with_suffix, "w") as f:
            data = pickle.loads(self.parameters.to_plot.value)

            # filter out keys that are not json-able
            for k, v in list(data.items()):
                if isinstance(v, np.ndarray):
                    data[k] = v.tolist()

            json.dump(data, f)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.


from typing import Optional

from linien_client.device import Device, add_device, update_device
from linien_gui.config import UI_PATH
from PyQt5 import QtWidgets, uic


class NewDeviceDialog(QtWidgets.QDialog):
    deviceName: QtWidgets.QLineEdit
    host: QtWidgets.QLineEdit
    username: QtWidgets.QLineEdit
    password: QtWidgets.QLineEdit
    port: QtWidgets.QSpinBox
    explainHostLabel: QtWidgets.QLabel

    def __init__(self, device: Optional[Device] = None) -> None:
        super(NewDeviceDialog, self).__init__()
        uic.loadUi(UI_PATH / "new_device_dialog.ui", self)

        if device is None:
            self.is_new_cevice = True
            self.device = Device()  # create a new empty device
        else:
            self.is_new_cevice = False
            self.device = device
            self.explainHostLabel.setVisible(False)

        self.deviceName.setText(self.device.name)
        self.host.setText(self.device.host)
        self.username.setText(self.device.username)
        self.password.setText(self.device.password)
        self.port.setValue(self.device.port)

    def add_new_device(self):
        self.device.name = self.deviceName.text()
        self.device.host = self.host.text()
        self.device.username = self.username.text()
        self.device.password = self.password.text()
        self.device.port = self.port.value()

        if self.is_new_cevice:
            add_device(self.device)
        else:
            update_device(self.device)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_gui.utils import get_linien_app_instance
from PyQt5 import QtCore, QtWidgets


class RightPanel(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs) -> None:
        super(RightPanel, self).__init__(*args, **kwargs)
        self.app = get_linien_app_instance()
        self.app.connection_established.connect(self.on_connection_established)
        QtCore.QTimer.singleShot(100, self.ready)

    def ready(self) -> None:
        self.main_window = self.app.main_window
        self.main_window.closeButton.clicked.connect(self.app.quit)
        self.main_window.shutdownButton.clicked.connect(self.shutdown_server)
        self.main_window.openDeviceManagerButton.clicked.connect(
            self.open_device_manager
        )
        self.main_window.pid_parameter_optimization_button.clicked.connect(
            self.open_psd_window
        )

    def on_connection_established(self) -> None:
        self.parameters = self.app.parameters
        self.control = self.app.control

        self.parameters.autolock_running.add_callback(self.autolock_status_changed)
        self.parameters.optimization_running.add_callback(
            self.optimization_status_changed
        )
        self.parameters.lock.add_callback(self.enable_or_disable_panels)

        def highlight_psd_button(locked: bool) -> None:
            self.main_window.pid_parameter_optimization_button.setStyleSheet(
                "background: #00aa00;" if locked else ""
            )

        self.parameters.lock.add_callback(highlight_psd_button)

    def shutdown_server(self) -> None:
        self.app.shutdown()

    def open_psd_window(self) -> None:
        self.app.open_psd_window()

    def open_device_manager(self) -> None:
        self.app.open_device_manager()

    def autolock_status_changed(self, value: bool) -> None:
        if value:
            self.main_window.settings_toolbox.setCurrentWidget(
                self.main_window.lockingPanel
            )

        self.enable_or_disable_panels()

    def optimization_status_changed(self, value: bool) -> None:
        if value:
            self.main_window.settings_toolbox.setCurrentWidget(
                self.main_window.optimizationPanel
            )

        self.enable_or_disable_panels()

    def enable_or_disable_panels(self, *args) -> None:
        lock = self.parameters.lock.value
        autolock = self.parameters.autolock_running.value
        optimization = self.parameters.optimization_running.value

        def enable_panels(panel_names, condition: bool) -> None:
            for panel_name in panel_names:
                getattr(self.main_window, panel_name).setEnabled(condition)

        enable_panels(("generalPanel",), not autolock and not optimization and not lock)
        enable_panels(
            ("modSpectroscopyPanel", "viewPanel", "lockingPanel"), not optimization
        )
        enable_panels(("optimizationPanel",), not autolock and not lock)
